# Added this file on the theory the import machinery requires it
# to access stuff further down

# in the next line, the first bifs is this package
# bifscore is the file bifscore.py within the package.
# and the last BIFS is a class definition from that file.
# Since BIFS is not a module, I don't think __all__ will work with it.
from bifs.bifscore import BIFS
from bifs import pset_dialogs



# Dialog window set up for BIFS parameters

from PyQt5 import QtCore, QtWidgets

class Param_Fourier_Space_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Parameter Space Function")
        Dialog.resize(200, 200)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.param_spaceSetUp = QtWidgets.QLabel("K-Space Parameter Function", self)
        self.param_spaceBox = QtWidgets.QComboBox(self)
        # self.param_spaceBox.setCurrentIndex(self.parent().param_func)
        self.param_spaceBox.addItems(self.parent().mybifs.param_func_choices)
        self.param_spaceBox.setCurrentIndex(self.parent().mybifs.param_func_choices.index(self.parent().mybifs.param_func_type))
        self.decay_label = QtWidgets.QLabel('Decay Exponent:', self)
        self.decay = QtWidgets.QLineEdit(str(self.parent().mybifs.decay))
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.param_spaceSetUp)
        self.layout.addWidget(self.param_spaceBox)
        self.layout.addWidget(self.decay_label)
        self.layout.addWidget(self.decay)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

class Prior_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Prior Distribution")
        Dialog.resize(200, 200)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.priorSetUp = QtWidgets.QLabel("Prior Distribution", self)
        self.priorBox = QtWidgets.QComboBox(self)
        # self.priorBox.setCurrentIndex(self.parent().param_func)
        self.priorBox.addItems(self.parent().mybifs.prior_choices)
        self.priorBox.setCurrentIndex(self.parent().mybifs.prior_choices.index(self.parent().mybifs.prior))
        self.scale_label = QtWidgets.QLabel('Scale:', self)
        self.scale = QtWidgets.QLineEdit(str(self.parent().mybifs.prior_scale))
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.priorSetUp)
        self.layout.addWidget(self.priorBox)
        self.layout.addWidget(self.scale_label)
        self.layout.addWidget(self.scale)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

class Likelihood_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Likelihood Distribution")
        Dialog.resize(200, 200)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.likelihoodSetUp = QtWidgets.QLabel("Likelihood Distribution", self)
        self.likelihoodBox = QtWidgets.QComboBox(self)
        # self.priorBox.setCurrentIndex(self.parent().param_func)
        self.likelihoodBox.addItems(self.parent().mybifs.likelihood_choices)
        self.likelihoodBox.setCurrentIndex(self.parent().mybifs.likelihood_choices.index(self.parent().mybifs.likelihood))
        self.scale_label = QtWidgets.QLabel('Scale:', self)
        self.scale = QtWidgets.QLineEdit(str(self.parent().mybifs.likelihood_scale))
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.likelihoodSetUp)
        self.layout.addWidget(self.likelihoodBox)
        self.layout.addWidget(self.scale_label)
        self.layout.addWidget(self.scale)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

class Slice3D_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Select 3D Slice to View")
        Dialog.resize(200, 200)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.slice_index_label = QtWidgets.QLabel('Slice Perpendicular to this Axis:', self)
        self.slice_index = QtWidgets.QComboBox()
        for i in range(3):
            self.slice_index.addItem(str(i), i)
        self.slice_index.setCurrentIndex(self.parent().mybifs.view3Dslice[0])
        #self.slice_index = QtWidgets.QLineEdit(str(self.parent().mybifs.view3Dslice[0]))
        self.slice_percent_label = QtWidgets.QLabel('Fraction Along Index:', self)
        self.slice_percent = QtWidgets.QLineEdit(str(self.parent().mybifs.view3Dslice[1]))
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.slice_index_label)
        self.layout.addWidget(self.slice_index)
        self.layout.addWidget(self.slice_percent_label)
        self.layout.addWidget(self.slice_percent)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

class AddBump_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Add K-Space Bump Filter")
        Dialog.resize(300, 300)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.addBumpKeySetUp = QtWidgets.QLabel("Add Bump", self)
        self.addBumpKeyBox = QtWidgets.QComboBox(self)
        self.addBumpKeyBox.addItems(self.parent().mybifs.bump_types)
        self.addBumpKeyBox.setCurrentIndex(self.parent().mybifs.bump_types.index(self.parent().mybifs.bump_default_type))
        self.Position_label = QtWidgets.QLabel('Position,fraction of Kmax:', self)
        self.Position = QtWidgets.QLineEdit()
        self.Amplitude_label = QtWidgets.QLabel('Amplitude,fraction of Pfuncmax:', self)
        self.Amplitude = QtWidgets.QLineEdit()
        self.Width_label = QtWidgets.QLabel('Width,fraction of Kmax:', self)
        self.Width = QtWidgets.QLineEdit()
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.addBumpKeySetUp)
        self.layout.addWidget(self.addBumpKeyBox)
        self.layout.addWidget(self.Position_label)
        self.layout.addWidget(self.Position)
        self.layout.addWidget(self.Amplitude_label)
        self.layout.addWidget(self.Amplitude)
        self.layout.addWidget(self.Width_label)
        self.layout.addWidget(self.Width)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

class DeleteBump_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Delete K-Space Bump Filter")
        Dialog.resize(300, 200)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setGeometry(QtCore.QRect(150, 250, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.deleteBumpKeySetUp = QtWidgets.QLabel("Delete Bump", self)
        self.deleteBumpKeyBox = QtWidgets.QComboBox(self)
        self.deleteBumpKeyBox.addItems(self.parent().bumplist)
        self.deleteBumpKeyBox.setCurrentIndex(self.parent().bumplist.index(self.parent().bumplist[0]))
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        self.layout.addWidget(self.deleteBumpKeySetUp)
        self.layout.addWidget(self.deleteBumpKeyBox)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)



# Prior based on empirical distribution of data

from bifs.priors.AbstractPrior import *

class EmpiricalPrior(AdjustmentPrior):
    """Empirical Prior
    Since this is an AdjustmentPrior one can add bumps to it.
    The interpretation of scale is different than for FunctionalPriors; here
    it is a multiplicative factor on the empirical sd.
    """
    def __init__(self, data,  basis, bvec, scale=1.0, scale_origin=10.**7):
        """initialize self with data and optional scale.
        data["mean"] is a numpy array
        data["sd"] is a numpy array
        scale will be multiplied by the sd to get the final value
        Note that scale_origin continues to be used as is, without multiplication
        """
        super().__init__(basis, bvec, scale, scale_origin)
        self._emean = data["mean"]
        self._esd = data["sd"]
        self._iorigin = tuple(0 for x in range(len(self._emean.shape)))
        
        # Assure all caches exist
        self._markDirty()

    def name(self):
        return "Empirical"

    def mean(self):
        if self._mean is None:
            self._mean = self._compute_mean()
        return self._mean

    def sd(self):
        if self._sd is None:
            self._sd = self._compute_sd()
        return self._sd

    def _compute_mean(self):
        mean = self._emean
        return self._adjustMean(mean)

    def _compute_sd(self):
        sd = self._esd*self._scale
        sd[self._iorigin] = self._scale_origin
        return sd

    def var(self):
        # it might be faster not to cache this at all
        if self._var is None:
            s = self.sd()
            self._var = s*s
        return self._var

    def _markDirty(self):
        super()._markDirty()
        self._mean = None
        self._sd = None
        self._var = None


###                       BIFS GUI                 ###
###                                                ###
###        This file contains the GUI interface    ###
###        for the BIFS package                    ###
###                                                ###

import cProfile, pstats
try:
    # new in 3.7
    from pstats import SortKey
    newProfile = True
except:
    newProfile = False

from PyQt5 import QtGui, QtWidgets, QtCore
from PyQt5.QtWidgets import QMessageBox
# docs for PySide2 say import Slot would do for next line
from PyQt5.QtCore import pyqtSlot as Slot
# next is not currently needed but for future-proofing
# also likely different in PySide2
from PyQt5.QtCore import pyqtSignal as Signal
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib.colors import NoNorm 
import matplotlib.pyplot as plt
from pylab import *
import numpy as np
from scipy import misc
import bifs
from bifs.bifsexception import *
import jsonpickle
import os
import itertools
import re

from bifs.pset_dialogs import Param_Fourier_Space_Dialog,Prior_Dialog
from bifs.pset_dialogs import Likelihood_Dialog,Slice3D_Dialog
from bifs.pset_dialogs import AddBump_Dialog,DeleteBump_Dialog

import logging
# the default log does not display anything graphically, which is not ideal. RB
log = logging.getLogger("Bifs")

# gastly hack
# but currently if this is run in debug mode it has a different working directory than
# if run from command line.  To avoid problems, hard code whole path.  Path removed for public release.
# Empirical Prior file
EPFILE = r"ep1.npz"


class MainWindow(QtWidgets.QMainWindow):
    """

    Class that generates the main BIFS GUI window

    Methods corresponding to QtActions are wrapped with @catcher; nothing else should be.
 
    """
    def __init__(self, app):
        super(MainWindow, self).__init__()
        self.main_widget = QtWidgets.QWidget(self)
        # Initialize bifs object
        self.mybifs = bifs.BIFS()
        self.myapp = app
        # use self.myapp.quit() for successful completion or
        # self.myapp.exit(code) for errors
        # No sys.exit is necessary.
        self.filename = None
        self.didMAP = False
        self.setWindowTitle("Bayesian Imaging in Fourier Space")
        self.fig = Figure()
        plt.rcParams["axes.grid"] = False # turn off grid lines for images
        plt.rcParams["xtick.color"] = (1,1,1,0)
        plt.rcParams["ytick.color"] = (1,1,1,0)
        self.ax1 = self.fig.add_subplot(221,adjustable='box')
        self.ax2 = self.fig.add_subplot(222,adjustable='box')
        self.ax3 = self.fig.add_subplot(223,adjustable='box')
        self.ax4 = self.fig.add_subplot(224,adjustable='box')
        self.axes=[self.ax1, self.ax2, self.ax3, self.ax4]
        self.canvas = FigureCanvas(self.fig)

        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.Expanding, 
                                   QtWidgets.QSizePolicy.Expanding)
        self.canvas.updateGeometry()
        self.layout = QtWidgets.QGridLayout(self.main_widget)
        self.layout.addWidget(self.canvas,0,0,5,5)
        self.createActions()
        self.createMenus()

        # left, top, width, height
        self.setGeometry(30,30,850,850)
        self.setCentralWidget(self.main_widget)
        self._signal_setup()
        self.show()
        self.update()

    def _signal_setup(self):
        "connect myself to current mybifs object"
        self.mybifs.image_loaded.connect(self.set_image_loaded)
        self.mybifs.image_unloaded.connect(self.set_image_unloaded)

    @Slot()
    def set_image_loaded(self):
        #enable menu
        self.loadEmpiricalPriorAct.setEnabled(True)  # must have loaded image
        self.loadEmpiricalPriorAct.setText("&Load Empirical Prior")


    @Slot()
    def set_image_unloaded(self):
        # disable menu
       self.loadEmpiricalPriorAct.setEnabled(False)  # must have loaded image
       self.loadEmpiricalPriorAct.setText("&Load Empirical Prior (requires loaded image)")

    @catcher
    def getImage(self):
        self.getImage_real()


    @catcher
    def getEmpiricalPrior(self):
        """

        Prompt for root directory of images.
        Scan each, FFT, and get mean and sd of values in k-space
        """
        docDir = QtCore.QStandardPaths.standardLocations(QtCore.QStandardPaths.DocumentsLocation)[0]
        topDirDlg = QtWidgets.QFileDialog(self, caption="Top Directory for Images",
                                         directory = docDir)
        topDirDlg.setFileMode(QtWidgets.QFileDialog.Directory)
        if topDirDlg.exec_():
            topDir = topDirDlg.selectedFiles()[0]
        else:
            return

        ## actual computation
        if self._scanImages(topDir):
            self._statsPost()
        else:
            # something funky in images
            pass
        return

    def _scanImages(self, rootDir):
        """
        Read in all the images under rootDir
        Fourier transform them

        self._mbifs  is a collection of the resulting bifs objects

        If headers of images are inconsistent, report error and return False.
        Otherwise return True.

        May operate in parallel.
        """

        self.nImages = 0
        benchmarkHdr = None
        mismatch = set()  # holds keys that had a mismatch
        reFile = re.compile(r"^mniwSUVR_.*\.nii(\.gz)?$", re.I)

        for root, dirs, files in os.walk(rootDir):
            # avoid our target case for whom we are trying to predict
            iKill = [ i for i, d in zip(itertools.count(), dirs) if d.find("10933")>=0]
            if iKill:
                nKill = 0
                for i in iKill:
                    i -= nKill
                    log.info("Skipping {}".format(dirs[i]))
                    del dirs[i-nKill]
                    nKill += 1
            # look for files to import
            if files:
                for f in files:
                    #if not f.endswith(".nii"):
                    if not reFile.match(f):
                        continue
                    if f.find("10933")>=0:
                        log.info("Skipping {}".format(f))
                        continue
                    self.nImages += 1
                    b = self.mybifs.copy_params()
                    b.load_image_file(os.path.join(root, f))
                    self._statsAccumulate(b.mod_image())
                    hdr = b.read_imfile.header
                    del b
                    if not benchmarkHdr:
                        # first header encountered
                        benchmarkHdr = hdr
                        # could not delete the following key
                        # it actually doesn't appear in the objects attributes
                        #del benchmarkHdr.__dict__['db_name']  # differences expected and no concern
                    else:
                        for key in benchmarkHdr:
                            if key == 'db_name':
                                continue

                            if key.startswith("scl_"):
                                # values were array(nan, dtype=float32) and I had no luck testing for them
                                # in various ways
                                continue
                            v1 = benchmarkHdr[key]
                            v2 = hdr[key]
                            if (v1 != v2).any():
                                mismatch.add(key)
        if mismatch:
            msgb = QMessageBox()
            msgb.setText("Warning: The following keys were not uniform in the files scanned: {}.".format(mismatch))
            msgb.setInformativeText("Do you want to proceed anyway?")
            msgb.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            msgb.setIcon(QMessageBox.Warning)
            msgb.setDefaultButton(QMessageBox.No)
            return msgb.exec() == QMessageBox.Yes
        return True

    def _statsAccumulate(self, m):
        """
        Accumulate running statistics on the values of m in different images
        self.nImages gives the current image number; it starts at 1.
        m is ordinarily the modulus, and must conform to numpy array protocols

        Updates self._mns and self._ss currently.
        """
        if self.nImages == 1:
            self._mns = m
            # ss will turn into a matrix later
            self._ss = 0.0
        else:
            lastdelta = m-self._mns
            self._mns += (lastdelta)/self.nImages
            # element by element multiplication in next line
            self._ss += lastdelta*(m-self._mns)

    def _statsPost(self):
        """
        Finalize computation of voxel by voxel statistics for all images.
        Call after all images have been seen.

        Results returned as arrays self.mns and self.sds.
        Also writes these results to disk.

        Note the assumption that all images have the same dimensions.
        Also, images should be aligned with each other for this to be meaningful

        """
        self.mns = self._mns
        # element by element square root
        self.sds = np.sqrt(self._ss/(self.nImages-1))
        del self._mns
        del self._ss
        np.savez(EPFILE, mean=self.mns, sd=self.sds)
        log.info("{} has means and sds for modulus of {} PET scans".format(EPFILE, self.nImages))
        #np.savez_compressed("ep1_compressed.npz", mean=mns, sd=sds)

    @catcher
    def loadEmpiricalPrior(self):
        """
        Load a previously computed empirical prior and use it as the prior for image reconstruction.

        getEmpiricalPrior scans through directories of images to build up the prior.
        Ultimately it saves the mean and sd of the modulus, as seen in the code immediately above.

        In contrast, this function reads those values that were written out and instructs
        mybifs to use those, rather than functional shortcuts, for the prior in reconstructing
        a particular image.
        """
        if not self.mybifs.image_file_loaded:
            # should never happen since action should be disabled
            QtWidgets.QMessageBox.information(self,"Empirical Prior", "You must load image using Load Initial Image... option from BIFS menu before loading the empirical prior.")
            return
        self.mybifs.load_empirical(EPFILE)

    def getImage_real(self):
        """
    
        function that calls the BIFS function load_image_file()
        to load an image file.  

        """
        # self.__init__()

        self.fileName = QtWidgets.QFileDialog.getOpenFileName(self, "Open File",
                                                     QtCore.QDir.currentPath())[0]
        # Qt docs say return value is a string, but it is tuple whose first element we want

        self.mybifs.load_image_file(self.fileName)
        self.show_initial_image()
        return


    @catcher
    def doMAP(self):
        """
    
        function that calls the BIFS function BIFS_MAP()
        to perform a MAP analysis.  

        """
        log.info("Performing k-space MAP estimation")
        # RB: original version of code created a new bifs instance.
        # I changed that to making one with copy_params.
        # Then I had it reload the image.  But under the current scheme that
        # will eliminate all customizations, which copy_params mostly does anyway.
        # Hopefully the original invalidity with mutliple image loads will no longer 
        # be possible in the new scheme.  Although loading a new image will destroy 
        # all customization in the new scheme.

        #self.mybifs.BIFS_MAP() should now happen implicitly
        self.show_post_proc_images()
        self.didMAP = True

        return
            
    def show_initial_image(self):
        """
    
        function that displays the initial image.  

        """
        if self.mybifs.image_file_loaded == True:
            self.ax1.clear()
            self.canvas.draw()
            self.ax1.set_title("Initial Image")
            if self.mybifs.imdim == 1:
                self.ax1.plot(self.mybifs.init_image()) 
            elif self.mybifs.imdim == 2:
                self.ax1.imshow(self.mybifs.init_image(),cmap = cm.Greys_r,norm=NoNorm())
            else: # assume for now that the only other possibility is 3D
                    # can change later
                slice_index = np.intp(np.round(self.mybifs.view3Dslice[1]*self.mybifs.init_image().shape[self.mybifs.view3Dslice[0]]))
                if self.mybifs.view3Dslice[0] == 0:
                    init_im_slice = self.mybifs.init_image()[slice_index,:,:]
                elif self.mybifs.view3Dslice[0] == 1:
                    init_im_slice = self.mybifs.init_image()[:,slice_index,:]
                elif self.mybifs.view3Dslice[0] == 2:
                    init_im_slice = self.mybifs.init_image()[:,:,slice_index]
                else:
                    raise BifsBadIndex("Sorry slice index needs to be one of 0,1,2")
                self.ax1.imshow(init_im_slice, cmap = cm.Greys_r)
            self.canvas.draw()
        return

        
    def show_post_proc_images(self):
        """
    
        function that displays the processed images.  

        """
        # Show initial and post BIFS k-spage images and final image
        self.ax1.set_title("Initial Image") # For some reason plt.cla()
                                            # clears plot 1 title
        self.ax2.clear()
        self.canvas.draw()
        self.ax2.set_title("Reconstructed Image")
        if self.mybifs.imdim == 1:
            self.ax2.plot(self.mybifs.final_image()) 
        elif self.mybifs.imdim == 2:
            self.ax2.imshow(self.mybifs.final_image(),cmap = cm.Greys_r)
        else: # assume for now that the only other possibility is 3D
                # can change later
            slice_index = np.intp(np.round(self.mybifs.view3Dslice[1]*self.mybifs.final_image().shape[self.mybifs.view3Dslice[0]]))
            if self.mybifs.view3Dslice[0] == 0:
                final_im_slice = self.mybifs.final_image()[slice_index,:,:]
            elif self.mybifs.view3Dslice[0] == 1:
                final_im_slice = self.mybifs.final_image()[:,slice_index,:]
            elif self.mybifs.view3Dslice[0] == 2:
                final_im_slice = self.mybifs.final_image()[:,:,slice_index]
            else:
                raise BifsBadIndex("Sorry slice index needs to be one of 0,1,2")
            self.ax2.imshow(final_im_slice, cmap = cm.Greys_r)
        self.canvas.draw()
                
        self.ax3.clear()
        self.canvas.draw()
        self.ax3.set_title("Initial K-Space Modulus Image")
        if self.mybifs.imdim == 1:
            showim1k = np.roll(np.roll(self.mybifs.mod_image(),self.mybifs.mod_image().shape[0]//2,0),1)
            self.ax3.plot(np.log(showim1k)) 
        elif self.mybifs.imdim == 2:
            showim1k = np.roll(np.roll(self.mybifs.mod_image(),self.mybifs.mod_image().shape[0]//2,0),self.mybifs.mod_image().shape[1]//2,1)
            self.ax3.imshow(np.log(showim1k),cmap = cm.Greys_r)
        else: # assume for now that the only other possibility is 3D
                # can change later
            slice_index = np.intp(np.round(self.mybifs.view3Dslice[1]*self.mybifs.mod_image().shape[self.mybifs.view3Dslice[0]]))
            if self.mybifs.view3Dslice[0] == 0:
                init_mod_im_slice = self.mybifs.mod_image()[slice_index,:,:]
            elif self.mybifs.view3Dslice[0] == 1:
                init_mod_im_slice = self.mybifs.mod_image()[:,slice_index,:]
            elif self.mybifs.view3Dslice[0] == 2:
                init_mod_im_slice = self.mybifs.mod_image()[:,:,slice_index]
            else:
                raise BifsBadIndex("Sorry slice index needs to be one of 0,1,2")

            showim1k = np.roll(np.roll(init_mod_im_slice,init_mod_im_slice.shape[0]//2,0),init_mod_im_slice.shape[1]//2,1)
            self.ax3.imshow(np.log(showim1k), cmap = cm.Greys_r)
        self.canvas.draw()

        self.ax4.clear()
        self.canvas.draw()
        self.ax4.set_title("BIFS K-Space Image")
        if self.mybifs.imdim == 1:
            showim2k = np.roll(np.roll(self.mybifs.bifsk_image(),self.mybifs.bifsk_image().shape[0]//2,0),1)
            self.ax4.plot(np.log(showim2k)) 
        elif self.mybifs.imdim == 2:
            showim2k = np.roll(np.roll(self.mybifs.bifsk_image(),self.mybifs.bifsk_image().shape[0]//2,0),self.mybifs.bifsk_image().shape[1]//2,1)
            self.ax4.imshow(np.log(showim2k),cmap = cm.Greys_r)
        else: # assume for now that the only other possibility is 3D
                # can change later
            slice_index = np.intp(np.round(self.mybifs.view3Dslice[1]*self.mybifs.bifsk_image().shape[self.mybifs.view3Dslice[0]]))
            if self.mybifs.view3Dslice[0] == 0:
                bifs_mod_im_slice = self.mybifs.bifsk_image()[slice_index,:,:]
            elif self.mybifs.view3Dslice[0] == 1:
                bifs_mod_im_slice = self.mybifs.bifsk_image()[:,slice_index,:]
            elif self.mybifs.view3Dslice[0] == 2:
                bifs_mod_im_slice = self.mybifs.bifsk_image()[:,:,slice_index]
            else:
                raise BifsBadIndex("Sorry slice index needs to be one of 0,1,2")

            showim2k = np.roll(np.roll(bifs_mod_im_slice,bifs_mod_im_slice.shape[0]//2,0),bifs_mod_im_slice.shape[1]//2,1)
            self.ax4.imshow(np.log(showim2k), cmap = cm.Greys_r)
        self.canvas.draw()
        return
    
    @catcher
    def saveCurrent(self):
        """
    
        function that calls the BIFS function save_results()
        to save the results of analysis.  

        """
        if not self.mybifs.final_image():
            # probably the test will throw an error rather than return
            QtWidgets.QMessageBox.information(self,"Save Current Results","No Results to Output Yet; Probably need to run - Get MAP Estimate Image - first")
        else:
            log.info("Saving Current Results")
            self.mybifs.save_results()

    @catcher
    def loadPrevious(self):
        """
    
        function that loads the results of a previous analysis
        saved via the BIFS function save_results().  

        """        
        fileName = QtWidgets.QFileDialog.getOpenFileName(self, "Load Existing Paramter File",QtCore.QDir.currentPath())[0]
        # Check if it's a bifs parameter file
        if fileName.split('.')[1] == 'bifspout':
            bifs_handle = open(fileName,"r")
            bifsj = bifs_handle.read()
            bifs_handle.close()
            self.mybifs = jsonpickle.decode(bifsj)
            self._signal_setup()
            # Show intial image
            self.show_initial_image()
            # Show initial and post BIFS k-spage images and final image
            self.show_post_proc_images()                    
        else:
            QtWidgets.QMessageBox.information(self, "Parameter File Loader", "Cannot load {}. Doesn't look like a bifs parameter file".format(fileName))
            return
        return
    @catcher    
    def close(self):
        self.myapp.quit()
        
    @catcher
    def zoomIn(self):
        self.scaleImage(1.25)

    @catcher
    def zoomOut(self):
        self.scaleImage(0.8)

    @catcher
    def normalSize(self):
        self.imageLabel.adjustSize()
        self.scaleFactor = 1.0

    @catcher
    def fitToWindow(self):
        fitToWindow = self.fitToWindowAct.isChecked()
        self.scrollArea.setWidgetResizable(fitToWindow)
        if not fitToWindow:
            self.normalSize()

        self.updateActions()

    def createActions(self):

        self.setParamFuncAct = QtWidgets.QAction("&Parameter Space Function", self, triggered=self.setParamFunc)

        self.setPriorAct = QtWidgets.QAction("&Prior Distribution", self, triggered=self.setPrior)

        self.setLikelihoodAct = QtWidgets.QAction("&Likelihood Distribution", self, triggered=self.setLikelihood)

        self.setAddBumpAct =  QtWidgets.QAction("&Add K-Space Bump", self, triggered=self.setAddBump)

        self.setDeleteBumpAct =  QtWidgets.QAction("&Delete K-Space Bump", self, triggered=self.setDeleteBump)
        
        self.set3DSliceAct = QtWidgets.QAction("&3D Slice to View", self, triggered=self.set3DSlice)
        
        self.getImageAct = QtWidgets.QAction("&Load Initial Image...", self, triggered=self.getImage)

        self.getEmpiricalPriorAct = QtWidgets.QAction("&Empirical Prior...", self, triggered=self.getEmpiricalPrior)

        self.loadEmpiricalPriorAct = QtWidgets.QAction("&Load Empirical Prior (requires loaded image)", self, triggered=self.loadEmpiricalPrior)
        self.loadEmpiricalPriorAct.setEnabled(False)  # must have loaded image

        self.doMapAct = QtWidgets.QAction("&Get MAP Estimate Image...", self,triggered=self.doMAP)

        self.saveCurrentAct = QtWidgets.QAction("&Save Current Results...", self,triggered=self.saveCurrent)

        self.loadPreviousAct = QtWidgets.QAction("&Load Previous Results...", self,triggered=self.loadPrevious)
        
        self.exitAct = QtWidgets.QAction("&Exit", self, shortcut="Ctrl+Q",
                triggered=self.close)

        self.zoomInAct = QtWidgets.QAction("Zoom &In (25%)", self,
                shortcut="Ctrl++", enabled=False, triggered=self.zoomIn)

        self.zoomOutAct = QtWidgets.QAction("Zoom &Out (25%)", self,
                shortcut="Ctrl+-", enabled=False, triggered=self.zoomOut)

        self.normalSizeAct = QtWidgets.QAction("&Normal Size", self,
                shortcut="Ctrl+S", enabled=False, triggered=self.normalSize)

        self.fitToWindowAct = QtWidgets.QAction("&Fit to Window", self,
                enabled=False, checkable=True, shortcut="Ctrl+F",
                triggered=self.fitToWindow)

        # self.aboutAct = QtWidgets.QAction("&About", self, triggered=self.about)

        # self.aboutQtAct = QtWidgets.QAction("&About Qt", self,
        #        triggered=QtWidgets.qApp.aboutQt)

    def createMenus(self):
        self.bifsMenu = QtWidgets.QMenu("&BIFS Operations", self)
        self.bifsMenu.addAction(self.getImageAct)
        self.bifsMenu.addAction(self.getEmpiricalPriorAct)
        self.bifsMenu.addAction(self.doMapAct)
        self.bifsMenu.addAction(self.saveCurrentAct)
        self.bifsMenu.addAction(self.loadPreviousAct)
        self.bifsMenu.addSeparator()
        self.bifsMenu.addAction(self.exitAct)

        self.paramMenu = QtWidgets.QMenu("&Parameter Set", self)
        self.paramMenu.addAction(self.loadEmpiricalPriorAct)
        self.paramMenu.addAction(self.setParamFuncAct)
        self.paramMenu.addAction(self.setPriorAct)
        self.paramMenu.addAction(self.setLikelihoodAct)
        self.paramMenu.addAction(self.setAddBumpAct)
        self.paramMenu.addAction(self.setDeleteBumpAct)
        self.paramMenu.addAction(self.set3DSliceAct)
        self.paramMenu.addSeparator()
        self.paramMenu.addAction(self.exitAct)
        
        self.viewMenu = QtWidgets.QMenu("&View", self)
        self.viewMenu.addAction(self.zoomInAct)
        self.viewMenu.addAction(self.zoomOutAct)
        self.viewMenu.addAction(self.normalSizeAct)
        self.viewMenu.addSeparator()
        self.viewMenu.addAction(self.fitToWindowAct)

        # self.helpMenu = QtWidgets.QMenu("&Help", self)
        # self.helpMenu.addAction(self.aboutAct)
        # self.helpMenu.addAction(self.aboutQtAct)
        
        self.menuBar().addMenu(self.bifsMenu)
        self.menuBar().addMenu(self.paramMenu)
        self.menuBar().addMenu(self.viewMenu)
        # self.menuBar().addMenu(self.helpMenu)

    def updateActions(self):
        self.zoomInAct.setEnabled(not self.fitToWindowAct.isChecked())
        self.zoomOutAct.setEnabled(not self.fitToWindowAct.isChecked())
        self.normalSizeAct.setEnabled(not self.fitToWindowAct.isChecked())
                    
    def adjustScrollBar(self, scrollBar, factor):
        scrollBar.setValue(int(factor * scrollBar.value()
                                + ((factor - 1) * scrollBar.pageStep()/2)))

    @catcher
    def setParamFunc(self):
        if self.mybifs.basis == "Fourier": # add else clauses as bases are added
            dialog = StartParam_Fourier_Space_Dialog(self)
            if dialog.exec_() == 1: # If OK button pressed
                self.mybifs.param_func_type = dialog.getFunc()
                self.mybifs.decay = dialog.getDecay()
        else:
            raise BifsBadInputs("Sorry, don't know how to handle {} basis yet.".format(self.mybifs.basis))

    @catcher
    def setPrior(self):
        dialog = StartPrior_Dialog(self)
        if dialog.exec_() == 1: # If OK button pressed
            self.mybifs.prior = dialog.getDist()
            self.mybifs.prior_scale = dialog.getScale()

    @catcher
    def setLikelihood(self):
        dialog = StartLikelihood_Dialog(self)
        if dialog.exec_() == 1: # If OK button pressed
            self.mybifs.likelihood = dialog.getDist()
            self.mybifs.likelihood_scale = dialog.getScale()

    @catcher
    def setAddBump(self):
        if not self.mybifs.image_file_loaded:
            QtWidgets.QMessageBox.information(self,"Add Bump", "Can't add bump without k-space structure, need to first load image using Load Initial Image... option from BIFS menu")
            return
        else:
            dialog = StartAddBump_Dialog(self)
            if dialog.exec_() == 1: # If OK button pressed
                get_key = dialog.getKey()
                get_position = dialog.getPosition()
                get_amplitude = dialog.getAmplitude()
                get_width = dialog.getWidth()
                self.mybifs.add_bump(get_key,get_position,get_amplitude,get_width)
                return

#                QtWidgets.QMessageBox.information(self, "Add Bump","Failed")
#                return

    @catcher
    def setDeleteBump(self):
        if self.mybifs.bumps:
            self.bumplist = []
            for keys in self.mybifs.bumps:
                self.bumplist.append(str(keys)+"  "+str(self.mybifs.bumps[keys][0]) + "  " + str(self.mybifs.bumps[keys][1]) + "  " + str(self.mybifs.bumps[keys][2]))
            dialog = StartDeleteBump_Dialog(self)
            if dialog.exec_() == 1: # If OK button pressed
                delstr = dialog.getDeleteString()
                delkey = str(delstr).split(" ")[0].strip("'[")
                del self.mybifs.bumps[delkey]
            return
        else:
            raise BifsBadInputs("No bumps to delete yet")
            return
    
    @catcher
    def set3DSlice(self):
        if self.mybifs.imdim != 3:
            raise BifsBadInputs("Sorry can only view slices for 3D data")
            return
        else:
            dialog = Start3DSlice_Dialog(self)
            if dialog.exec_() == 1: # If OK button pressed
                self.mybifs.view3Dslice[0] = dialog.getSliceIndex()
                self.mybifs.view3Dslice[1] = dialog.getSlicePercent()
                if self.didMAP:
                    self.show_initial_image()
                    self.show_post_proc_images()
                else:
                    if self.mybifs.image_file_loaded:
                        self.show_initial_image()
                return

            
class StartParam_Fourier_Space_Dialog(QtWidgets.QDialog, Param_Fourier_Space_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getDecay(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.decay))
        except:
            raise BifsBadParameter("Couldn't get decay value")
            
    def getFunc(self):
        try:
            return str(self.param_spaceBox.currentText())
        except:
            raise BifsBadParameter("Couldn't get parameter function")

class StartPrior_Dialog(QtWidgets.QDialog, Prior_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getScale(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.scale))
        except:
            raise BifsBadParameter("Couldn't get scale value")
            
    def getDist(self):
        try:
            return str(self.priorBox.currentText())
        except:
            raise BifsBadParameter("Couldn't get prior distribution")
            
class StartLikelihood_Dialog(QtWidgets.QDialog, Likelihood_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getScale(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.scale))
        except:
            raise BifsBadParameter("Couldn't get scale value")
            
    def getDist(self):
        try:
            return str(self.likelihoodBox.currentText())
        except:
            raise BifsBadParameter("Couldn't get likelihood distribution")

class Start3DSlice_Dialog(QtWidgets.QDialog, Slice3D_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getSliceIndex(self):
        try:
            # currentIndex should yield same value as currentData, but the latter
            # seems more future-proof.  RB
            return np.int(self.slice_index.currentData())
        except:
            raise BifsBadParameter("Couldn't get slice index")
            
    def getSlicePercent(self):
        # the value is actually a fraction between 0 and 1
        try:
            return float(QtWidgets.QLineEdit.text(self.slice_percent))
        except:
            raise BifsBadParameter("Couldn't get slice percent")

class StartAddBump_Dialog(QtWidgets.QDialog, AddBump_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getKey(self):
        try:
            return str(self.addBumpKeyBox.currentText())
        except:
            raise BifsBadParameter("Couldn't get bump key")
            
    def getPosition(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.Position))
        except:
            raise BifsBadParameter("Couldn't get bump position")
            
    def getAmplitude(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.Amplitude))
        except:
            raise BifsBadParameter("Couldn't get bump position")
            
    def getWidth(self):
        try:
            return float(QtWidgets.QLineEdit.text(self.Width))
        except:
            raise BifsBadParameter("Couldn't get bump width")
            
class StartDeleteBump_Dialog(QtWidgets.QDialog, DeleteBump_Dialog):
    def __init__(self,parent=MainWindow):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)

    def getDeleteString(self):
        try:
            return str(self.deleteBumpKeyBox.currentText())
        except:
            raise BifsBadParameter("Couldn't get bump function information")

def launch():
    "provided as a possible entry point"
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow(app)
    app.exec()  # for Python2 and Qt4 sys.exit(app.exec_())

if __name__ == '__main__':
    launch()



"""
This module defines Exception classes to indicate errors or warnings and functions to
install global handlers for those exceptions.  Global handlers are important since
exceptions may be raised from code executing inside the Qt event loop.

If used from a Qt program pops up a message box with exception info.  
Perhaps too much info (full traceback)

Exceptions are also logged to sys.stdout.

Exceptions
----------
All classes are subclasses of BifsException.  But note the code may raise other kinds of
exceptions.   BifsFatal is a base class for exceptions that should terminate the program, and they
always include an error code with which to terminate.  Other exceptions will typically function
as notifications.

Logging
-------

Goes to a logger named "Bifs".

Useage
------

In your main program
====================
import bifs.bifsexception
# will activate the error handlers
# if you need to use the exceptions or the log, use the patterns below instead.

Other modules
=============
from bifs.bifsexception import *
# and then raise any of the exceptions
# if you need to log something
import logging
log = logging.getLogger("Bifs")

GUI Wrapper for Errors
=======================
To avoid silent failures when running in a GUI, we want to catch all Exceptions and pop up
a  window with the relevant information.  Errors are also logged to the console.

Somewhat unfortunately, there are 2 mechanisms for doing this.

First, this module puts a hook in the global exception handler; this should catch everything.
However, an exception that makes it to that hook has already failed out of the main event
loop of the application.  Python's exception semantics means the app has failed and can't be
resumed.

To avoid that, we also define a function wrapper `@catcher` that goes around individual functions.
This catches errors while inside the event loop, allowing the user to decide
whether to continue with the application.  Use it like this:

@catcher
def myfunction(self):

The function wrapped must take a single argument only; typically it is a class method with `self` as 
the only argument.  The function must not return a value.

Why the restrictions on the function to be wrapped?
  No return value: The exception means the function didn't complete, and the exception handler
    doesn't know what return value(s) would be appropriate.  So it doesn't make sense to use it
    when a value needs to be returned.
  Single argument only: The anticipated use if for QtActions implemented as methods with only
    self as an argument (and no return value).
  Single argument only: There is a technical limitation in functools.wraps; consider
  ```python
  def mywrapper(fn):
    @wraps(fn)
    def _wrapper(*args, **kwds):
        return fn(*args, **kwds)
    return _wrapper

  class Foo:
    @mywrapper
    def bar(self):
        # do stuff
  ```
  `Foo.bar` will appear to have the signature `*args, **kwds` despite the fact that the original
  bar had only the argument `self` and `@wraps` is advertised as preserving the characteristics
  of the original function.  This means, in particular, that Qt will send 2 arguments to the
  wrapped function, which will cause an error when `fn(*args, **kwds)` attempts to invoke the original `bar`.

  The decorator package, in particular its decorator member, supposedly does a better job preserving the original
  signature.  But after reading https://github.com/micheles/decorator/blob/master/docs/documentation.md a few times
  I still wasn't sure how it was to be used.  Since it's an extra dependency, and since my needs seem to
  all be met with the single argument case, I just solved that problem.  RB


Implementation Note
-------------------
Python's import machinery should assure the code here is executed only once even if multiple
modules import this module.  For that to work you must use `import` rather than other strategies to access
the code.

Future Directions
-----------------
* ability to suppress traceback in dialog box
* separate the general and GUI-specific parts
* more use of logging

Credits
-------
Based on code from Tim Lehr https://timlehr.com/python-exception-hooks-with-qt-message-box/
He indicated no explicit license but did write, just below it
"Just paste it somewhere in one of your modules"

Modifications by Ross Boylan <ross.boylan@ucsf.edu> under GPL-3 or BSD-ish license for the whole package.

"""

__all__ = ("catcher", "BifsException", "BifsBadParameter", "BifsBadIndex", "BifsLookupFailure", "BifsBadInputs")


from functools import wraps
import sys
import traceback
import logging
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import pyqtSignal as Signal

def catcher(function):
    """A decorator to catch exceptions in raised while running function

    ONLY USE THIS TO WRAP FUNCTIONS OF A SINGLE ARGUMENT WITH NO RETURN VALUE
    Since intended use is for instance methods, the inner function uses self as the
    name of the single argument.
    """
    @wraps(function)
    def wrapper(self):
        try:
            return function(self)
        except:
            exc_info = sys.exc_info()
            exc_type, exc_value, exc_traceback = exc_info
            if issubclass(exc_type, KeyboardInterrupt):
                # ignore keyboard interrupt to support console applications
                # Maybe I could just re-raise the exception? RB
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
            else:
                log_msg = '\n'.join([''.join(traceback.format_tb(exc_traceback)),
                                     '{0}: {1}'.format(exc_type.__name__, exc_value)])
                if issubclass(exc_type, BifsException):
                    log_msg = str(exc_value)+"\nProceed with caution.\n"+log_msg
                    log.warn(log_msg)
                else:
                    log.critical("Uncaught exception:\n {0}".format(log_msg), exc_info=exc_info)

                # Unlike the code further down, this takes no steps to
                # assure execution in main thread.  Probably could cause trouble if there were
                # multiple threads.
                show_exception_box(log_msg)
    return wrapper

# basic logger functionality
log = logging.getLogger("Bifs")
handler = logging.StreamHandler(stream=sys.stdout)
log.addHandler(handler)

class BifsException(Exception):
    pass

class BifsBadParameter(BifsException):
    pass

class BifsBadIndex(BifsException):
    pass

class BifsLookupFailure(BifsException):
    pass

class BifsBadInputs(BifsException):
    pass

def show_exception_box(log_msg):
    """Checks if a QApplication instance is available and shows a messagebox with the exception message. 
    If unavailable (non-console application), log an additional notice.
    """
    if QtWidgets.QApplication.instance() is not None:
            errorbox = QtWidgets.QMessageBox()
            errorbox.setText("Oops. An unexpected error occured:\n{0}".format(log_msg))
            errorbox.exec_()
    else:
        log.debug("No QApplication instance available.")
 
class UncaughtHook(QtCore.QObject):
    _exception_caught = Signal(object)
 
    def __init__(self, *args, **kwargs):
        super(UncaughtHook, self).__init__(*args, **kwargs)

        # this registers the exception_hook() function as hook with the Python interpreter
        sys.excepthook = self.exception_hook

        # connect signal to execute the message box function always on main thread
        self._exception_caught.connect(show_exception_box)
 
    def exception_hook(self, exc_type, exc_value, exc_traceback):
        """Function handling uncaught exceptions.
        It is triggered each time an uncaught exception occurs. 
        """
        if issubclass(exc_type, KeyboardInterrupt):
            # ignore keyboard interrupt to support console applications
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
        else:
            exc_info = (exc_type, exc_value, exc_traceback)
            log_msg = '\n'.join([''.join(traceback.format_tb(exc_traceback)),
                                 '{0}: {1}'.format(exc_type.__name__, exc_value)])
            if issubclass(exc_type, BifsException):
                log_msg = str(exc_value)+"\nProceed with caution.\n"+log_msg
                log.warn(log_msg)
            else:
                log.critical("Uncaught exception:\n {0}".format(log_msg), exc_info=exc_info)

            # trigger message box show
            self._exception_caught.emit(log_msg)
 
# create a global instance of our class to register the hook
qt_exception_hook = UncaughtHook()



# A command line type example for using BIFS example

import numpy as np
import scipy as sp
from scipy import misc,stats
import random
from pylab import *
import matplotlib.pyplot as plt
import bifs
import bifs.bifs_util.util as bu

# Make 1D "image"
# Noise standard deviation in image space
noiseSD = 0.5
# Set y as a discrete set of values of even length
z = np.concatenate((np.zeros(14),np.arange(19)+1,np.zeros(7)+10,10-np.arange(10),np.zeros(14)))

y = z + stats.norm.rvs(size=len(z),loc=0.0,scale=noiseSD)

# Create mybifs BIFS object:
mybifs = bifs.BIFS()

# Can take a look at what functions and variables are available with, e.g.:
# dir(mybifs)

# Currently, loading an image resets everything else.  So start there.
# Load the image - note, typically just start here re. loading a noisy image
mybifs.load_image(y)

# Set a few things:
# Prior
mybifs.prior = "Gaussian" # Only possibility.
# We had a uniform prior at one point, but Karl Young says (from issue #4)
# I realized that doing “maximum likehood” analysis (the equivalent of using a uniform prior)
# was silly in our case as the data point (pixel value) is, by definition, the maximum of the
# likelihood distribution. So doing a fancy optimization that just returns the original image
# didn’t seem like the best use of cycles


# Lilelihood
mybifs.likelihood = "Gaussian" # Choices are currently: "Gaussian","Rician" 
# "Gaussian" is actually the default but again for illustration...

# Parameter Space Function
# Always set it via this function.
mybifs.set_prior_func_type("Linear Decay")
# Current choices are: "Inverse Power Decay","Banded Inverse Power Decay",
# "Linear Decay" with default "Inverse Power Decay",
# but again for illustration...

# Can check comments in bifs.py for description of other parametere to set



# Run BIFS making sure that the initial image is loaded

print("Running BIFS_MAP() on image")
# It is no longer necessary to call this explicitly.
# It will run implicitly whenever you request one of the images it creates.
mybifs.BIFS_MAP()

# Take a look at the current paramter function
bu.plot_param_func(mybifs)

# Look at the prior, liklelihood, and posterior at a voxel
bu.voxel_dist(mybifs,[mybifs.mod_image().shape[0]//2],do_plots=True)
  
# Plot the resulting "images"
# Initial noisy image
plt.subplot(221)
plt.axis('off')
plt.title("Initial Image")
plt.plot(mybifs.init_image())
  
# Initial K-Space Image
plt.subplot(223)
plt.axis('off')
plt.title("Initial K-Space Image")
showim1k = np.roll(np.roll(mybifs.mod_image(), mybifs.mod_image().shape[0]//2, 0), 1)
plt.plot(np.log(showim1k))
             
# Final K-Space Image after running BIFS
plt.subplot(224)
plt.axis('off')
plt.title("Final K-Space Image")
showim2k = np.roll(np.roll(mybifs.bifsk_image(),mybifs.bifsk_image().shape[0]//2,0),1)
plt.plot(np.log(showim2k))
             
# Final Image after running BIFS
plt.subplot(222)
plt.axis('off')
plt.title("Reconstructed Image")
plt.plot(mybifs.final_image())

plt.show()



# Use some functional form to simplify the prior
from bifs.priors.AbstractPrior import *

def param_func_choices():
    return ["Inverse Power Decay", "Banded Inverse Power Decay", "Linear Decay"]

class FunctionalPrior(AdjustmentPrior):
    """ Abstract class with common logic for all functional priors
    kdist is an array with distance from center of Fourier space but shifted so origin is at [0], [0,0] or [0,0,0]
            as appropriate for image dimension
    _iorgin holds a tuple with the origin index once dimensionality is set

    See AdjustmentPrior for other variables.

    Note that it is very important to do all setup of the mean before touching the sd.
    The only setters of the mean are internal to the class, but remember the sd depends on the mean.
    """
    def __init__(self, basis, kdist, bvec, scale = 0.0005, scale_origin = 10.**7):
        super().__init__(basis, bvec, scale, scale_origin)
        self.setkdist(kdist)
        # setkdist calls _markDirty

    def _markDirty(self):
        super()._markDirty()
        # see https://hynek.me/articles/hasattr/ for some reasons not to use hasattr for testing if a variable is present
        # Also https://stackoverflow.com/questions/24971061/python-hasattr-vs-getattr
        # So instead we set to None
        self._mean = None
        self._sd = None
        self._var = None

    def setkdist(self, kdist):
        self._kdist = kdist
        self._iorigin = tuple(0 for x in range(len(kdist.shape)))
        self._markDirty()

    def mean(self):
        if self._mean is None:
            self._mean = self._compute_mean()
        return self._mean

    def sd(self):
        if self._sd is None:
            self._sd = self._compute_sd()
        return self._sd

    def _compute_sd(self):
        """all functional forms currently share this scaling,
        which implies a strong prior"""
        sd = self.mean()*self._scale
        sd[self._iorigin] = self._scale_origin
        return sd

    def var(self):
        # it might be faster not to cache this at all
        if self._var is None:
            s = self.sd()
            self._var = s*s
        return self._var

class InversePowerDecayPrior(FunctionalPrior):
    """
    decay - float decay exponent for the inverse power paramter function
    """
    def __init__(self, basis, kdist, bvec = None, scale = 0.0005, scale_origin = 10.**7, decay = 0.5):
        if bvec is None:
            bvec = basis.bvec_ixsc
        super().__init__(basis, kdist, bvec, scale, scale_origin)
        self._decay = decay

    def name(self):
        return "Inverse Power Decay"

    def setdecay(self, decay):
        if decay != self.decay:
            self._decay = decay
            self._markDirty()

    def decay(self):
        return self._decay

    def _compute_mean(self):
        mean = self._basis.ixsc(self._bvec, self._kdist, self._decay)
        return self._adjustMean(mean)

class BandedInversePowerDecayPrior(InversePowerDecayPrior):
    """
    banded_cutoff - cutoff for banded inverse power k-space paramter function 
    """
    def __init__(self, basis, kdist, bvec = None, scale = 0.0005, scale_origin = 10.**7, decay = 0.5, banded_cutoff=50.):
        if bvec is None:
            bvec = basis.bvec_ixscbanded
        super().__init__(basis, kdist, bvec, scale, scale_origin, decay)
        self._banded_cutoff = banded_cutoff

    def name(self):
        return "Banded Inverse Power Decay"

    def setbanded_cutoff(self, bc):
        self._banded_cutoff = bc
        self._markDirty()

    def banded_cutoff(self):
        return self._banded_cutoff

    def _compute_mean(self):
        mean = self._basis.ixscbanded(self._bvec, self._kdist, self._decay, self._banded_cutoff)
        return self._adjustMean(mean)

class LinearDecayPrior(FunctionalPrior):
    """Simple linear decay.
    """
    def __init__(self, basis, kdist, bvec = None, scale = 0.0005, scale_origin = 10.**7):
        if bvec is None:
            bvec = basis.bvec_linsc
        super().__init__(basis, kdist, bvec, scale, scale_origin)

    def name(self):
        return "Linear Decay"

    def _compute_mean(self):
        mean = self._basis.linsc(self._bvec, self._kdist)
        return self._adjustMean(mean)


import itertools
import os
import re
import sys

# scans files to construct an empirical prior
from bifs import BIFS

# numpy >= 1.17
from numpy.random import Generator, PCG64
import numpy as np

class RunningMean:
    """Accepts values one at a time and computes the mean and sd of all values seen so far.
    The inputs are arrays, which must all have the same shape.  Mean and sd are accumulated
    separately for each cell in the array.
    """
    def __init__(self, sd=True):
        """If sd is false do not accumulate second moment.
        Clients should not request information related to the sd in that case.
        """
        self.n = 0
        self._second = sd  # as in second moment

    def observation(self, x):
        "x is an array-like object which is considered a single observation"
        self.n += 1
        if self.n == 1:
            self._mns = x
            if self._second:
                # ss will turn into a matrix later
                self._ss = 0.0
        else:
            lastdelta = x-self._mns
            self._mns += (lastdelta)/self.n
            if self._second:
                # element by element multiplication in next line
                self._ss += lastdelta*(x-self._mns)

    def mean(self):
        "return array of means so far"
        return self._mns

    def sd(self):
        "return array of sd so far"
        # element by element square root
        return np.sqrt(self._ss/(self.n-1))

class AbstractEmpiricalScanner:
    """ This class consumes a list of images and computes statistics on them.  Each statistic is computed separately for 
    each voxel, i.e. the result in the (2, 5) cell refers to all the (2, 5) cells in all the images (or their Fourier counterparts).
    All images must have the same dimensions, and they should be aligned with each other
    for the results to be meaningful.

    The mean and sd of the modulus is always accumulated; values for the phase can be requested as well, as can the correlations between the
    phase and modulus (again, at each point in Fourier space).

    Finally, one can request a sample of the original voxels in image space.

    Concrete classes provide particular ways to get images.  They then pass the images to _statsAccumulate and,
    optionally, _voxAccumulate (possibly different images for each) and call
    _post when done.  At that point, and only that point, are results available from self.modulus() and, if requested,
    self.phase(), self.corr(), and self.voxels().

    For backward compatility, self.mns, self.sds, and self.vox accessor return the mean and sd of self.modulus() and the voxels.
    Don't rely on that in new code.

    image_mask optionally indicates which areas of the image to ignore.
        It must be a boolean array with the same shape as image files.
        All voxels selected by image_mask are set to zero before doing BIFS processing.
        The mask applies to the original image NOT to the fourier space version, which will
        generally have non-0 values in the image_mask region.
        It is the subclass responsibility to implement these semantics.
        Note the "mask" here is not a mask in the numpy sense of a masked array, which
        concerns missing values.

    voxel sampling only considers the non-ignored regions, but the number sampled will be based on
        the total voxel count before masking.
    """
    def __init__(self, sampleFraction=0, seed=85792359, image_mask=None, phase=False, corr=False):
        """Setup for scan of images
        if sampleFraction is >0 (and it should be <=1) then that fraction of the image voxels will be retained.
        In that case, seed is used to set the random number generator.
        If phase is true, accumulate statistics on the phase as well as the modulus.
        If corr is true, accumulate statistics on the phase and its covariance with the modulus.
        Covariance is on  a cell by cell basis.
        """
        self.sampleFraction = sampleFraction
        self._modulus = RunningMean()
        if phase or corr:
            self._getPhase = True
            self._phase = RunningMean()
        if corr:
            self._getcorr = True
            self._xy = RunningMean(sd=False)
        if sampleFraction>0:
            self._voxels = []
            self._rg = Generator(PCG64(seed))
        self.masking = (image_mask is not None)
        if self.masking:
            self.image_mask = image_mask
            self.image_keep = np.logical_not(image_mask)
        self._benchmarkHdr = None  # checks for consistent headers
        self._mismatch = set()  # holds keys that had a mismatch
        self._bifs = BIFS()

    def modulus(self)->RunningMean:
        return self._modulus

    def phase(self)->RunningMean:
        return self._phase

    def corr(self):
        "Note we return the correlation matrix itself, not an accumulator"
        return (self._xy.mean()-self._modulus.mean()*self._phase.mean())/ \
            (self._modulus.sd()*self._phase.sd())


    def voxels(self):
        "return 1-d array sorted by intensity"
        return self._voxels

    def __getattr__(self, name):
        ## backward compatibility only
        if name == "mns":
            return self.modulus().mean()
        if name == "sds":
            return self.modulus().sd()
        if name == "vox":
            return self.voxels()
        raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, name))

    def _do_one(self, file):
        "file is a path-like object. Read it in and accumulate information"
        self._bifs.load_image_file(file)
        if self.masking:
            # dirty trick.  But doesn't invalidate anything else in _bifs.
            self._bifs._init_image[self.image_mask] = 0.0
        self._modulus.observation(self._bifs.mod_image())
        if self._getPhase:
            self._phase.observation(self._bifs.phase_image())
            if self._getcorr:
                # next multiplication is element by element
                self._xy.observation(self._bifs.phase_image()*self._bifs.mod_image())

        if self.sampleFraction>0:
            self._voxAccumulate(self._bifs.init_image())
        hdr = self._bifs.read_imfile.header
        if not self._benchmarkHdr:
            # first header encountered
            self._benchmarkHdr = hdr
            # could not delete the following key
            # it actually doesn't appear in the objects attributes
            #del benchmarkHdr.__dict__['db_name']  # differences expected and no concern
        else:
            for key in self._benchmarkHdr:
                if key == 'db_name':
                    continue
                if key.startswith("scl_"):
                    # values were array(nan, dtype=float32) and I had no luck testing for them
                    # in various ways
                    continue
                v1 = self._benchmarkHdr[key]
                v2 = hdr[key]
                if (v1 != v2).any():
                    self._mismatch.add(key)

    def _voxAccumulate(self, m):
        """accumulate voxel values.
        In the most likely case, the voxels are from image space while the empirical prior
        is from k-space.  So we provide seperate functions for the 2 values.
        Calling this is pointless unless sampleFraction>0.
        """
        # always base number sampled on the complete image size
        nSamp = int(m.size*self.sampleFraction)

        if self.masking:
            self._voxels.append(self._rg.choice(m[self.image_keep], nSamp))
        else:
            # m.ravel is not an acceptable first argument to choice
            # actually, it should have been np.ravel(m)
            # m.flatten and the mask selection above both create copies, unfortunately
            self._voxels.append(self._rg.choice(m.flatten(), nSamp))


    def _statsPost(self):
        """
        Finalize computation of voxel by voxel statistics for all images.
        Call after all images have been seen.

        Results returned as arrays self.mns and self.sds.
        """
        # currently handled by RunningMean instances automatically
        pass

    def _voxPost(self):
        """
        Finalize accumulated voxels.
        """
        if self.sampleFraction>0:
            self._voxels = np.concatenate(self._voxels)
            self._voxels.sort()


    def _post(self):
        "wrap up all processing"
        self._statsPost()
        self._voxPost()

    def nImages(self) -> int:
        "number of images processed so far = number of files read unless error"
        return self._modulus.n


class EmpiricalScanner(AbstractEmpiricalScanner):
    """Scan selected images on disk, ensuring they are alll compatible.

    topDir  path like object indicating where in the file system the scan should start
        all subdirectories will be scanned recursively unless they are excluded.
    matchFile  <String> regular expression for the file name of image files we want.
        Matching is on the file name only, not its full path.
    exclude     <String> optional regular expression.  Any directory matching this pattern is excluded.
        Any file that satisfies matchFile is excluded if it also matches exclude.
    ostr    A stream-like object that will receive routines notices of skipped files and statistics.

    See AbstractEmpiricalScanner for sampleFraction, seed and image_mask.

    The files are read in and converted to k-space.  We compute  the mean and sd of the k-space images,
    and optionally accumulate voxels from the original image.

    We also check that the headers are consistent.  This works for .nii files, and may or may not for others.
    """
    def __init__(self, sampleFraction=0, seed=85792359, topDir=".", matchFile="", exclude=None, image_mask=None, phase=False, corr=False, ostr=sys.stdout):
        super().__init__(sampleFraction, seed, image_mask, phase, corr)
        self._topDir = topDir
        self._matchRE = re.compile(matchFile, re.I)
        if exclude:
            self._excludeRE = re.compile(exclude, re.I)
        else:
            self._excludeRE = None
        self.go(ostr=ostr)

    def go(self, ostr=sys.stdout):
        """Actually perform the scan.
        Note this is triggered by object initialization.
        Repeated calls may not work.

        ostr is an output stream
        """
        for root, dirs, files in os.walk(self._topDir):
            if self._excludeRE:
                # avoid directories with our target case for whom we are trying to predict
                iKill = [ i for i, d in zip(itertools.count(), dirs) if self._excludeRE.search(d)]
                if iKill:
                    nKill = 0
                    for i in iKill:
                        i -= nKill
                        print("Skipping {}".format(dirs[i]), file=ostr)
                        del dirs[i-nKill]
                        nKill += 1
            # look for files to import
            if files:
                for f in files:

                    if not self._matchRE.search(f):
                        continue
                    if self._excludeRE:
                        if self._excludeRE.search(f):
                            print("Skipping {}".format(f), file=ostr)
                            continue
                    self._do_one(os.path.join(root, f))
        self._post()

class FeedScanner(AbstractEmpiricalScanner):
    """A scanner that accepts anything iterable as a list of file names to scan"""
    def __init__(self, files, sampleFraction=0, seed=85792359, image_mask=None, phase=False, corr=False, ostr=sys.stdout):
        super().__init__(sampleFraction, seed, image_mask, phase, corr)
        self._files = files
        self.go(ostr=ostr)

    def go(self):
        for f in self._files:
            self._do_one(f)
        self._post()



# A command line type example for using BIFS example

import numpy as np
import imageio
import random
from pylab import *
import matplotlib.pyplot as plt
import bifs
import bifs.bifs_util.util as bu

# 2D image
# Load image - standard Lena for now
im = imageio.imread('../tests/images/lena512.bmp')
im = np.asarray(im)
# 2D imag
noise_level = 1.5
noise = noise_level*(np.max(im) - np.min(im))*np.random.rand(im.shape[0],im.shape[1])

# Add noise:
noisy_im = im + noise

# Create mybifs BIFS object:
mybifs = bifs.BIFS()

# Can take a look at what functions and variables are available with, e.g.:
# dir(mybifs)

# Set a few things:
# Prior
mybifs.prior = "Gaussian" # Choices are currently: "Gaussian","Uniform"
# "Gaussian" is actually the default but for illustration...

# Lilelihood
mybifs.likelihood = "Gaussian" # Choices are currently: "Gaussian","Rician" 
# "Gaussian" is actually the default but again for illustration...

# Parameter Space Function
# mybifs.param_func_type = "Linear Decay"
# mybifs.bvec = np.array([0.1,500.])
mybifs.param_func_type = "Inverse Power Decay"
# mybifs.decay = 0.5
# Current choices are: "Inverse Power Decay","Banded Inverse Power Decay",
# "Linear Decay" with default "Inverse Power Decay",
# but again for illustration...

# Try adding a bump
# mybifs.bumps = {mybifs.bump_default_type:[0.5,0.05,0.1]}

# Can check comments in bifs.py for description of other parametere to set

# Load the image - note, typically just start here re. loading a noisy image
mybifs.load_image(noisy_im)

# Run BIFS MAP
print("Running BIFS_MAP() on image")
mybifs.BIFS_MAP()

# Take a look at the parameter function
# NOTE: For a reason I don't yet understand
# running this and/or the following utility
# functions before running mybifs.BIFS_MAP()
# causes mybifs.BIFS_MAP() to crash horribly
# with a string of errors like:
#
# Break on __THE_PROCESS_HAS_FORKED_AND_YOU_CANNOT_USE_THIS_
# COREFOUNDATION_FUNCTIONALITY___YOU_MUST_EXEC__() to debug.
# The process has forked and you cannot use this CoreFoundation
# functionality safely. You MUST exec().
#
# bu.plot_param_func(mybifs)
  
# Look at the prior, liklelihood, and posterior at a voxel
bu.voxel_dist(mybifs,[mybifs.mod_image().shape[0]//2,mybifs.mod_image().shape[1]//2],do_plots=True)

# Plot the resulting images
# Initial noisy image
plt.subplot(221)
plt.axis('off')
plt.title("Initial Image")
plt.imshow(mybifs.init_image(), cmap = cm.Greys_r)
  
# Initial K-Space Image
plt.subplot(222)
plt.axis('off')
plt.title("Initial K-Space Image")
showim1k = np.roll(np.roll(mybifs.mod_image(),mybifs.mod_image().shape[0]//2,0),mybifs.mod_image().shape[1]//2,1)
plt.imshow(np.log(showim1k), cmap = cm.Greys_r)
             
# Final K-Space Image after running BIFS
plt.subplot(224)
plt.axis('off')
plt.title("Initial K-Space Image")
showim2k = np.roll(np.roll(mybifs.bifsk_image(),mybifs.bifsk_image().shape[0]//2,0),mybifs.bifsk_image().shape[1]//2,1)
plt.imshow(np.log(showim2k), cmap = cm.Greys_r)
             
# Final Image after running BIFS
plt.subplot(223)
plt.axis('off')
plt.title("Reconstructed Image")
plt.imshow(mybifs.final_image(),cmap = cm.Greys_r)

plt.show()



# A command line type example for using BIFS example

import numpy as np
import imageio
import random
from pylab import *
import matplotlib.pyplot as plt
import bifs
import bifs.bifs_util.util as bu

# 3D image
# Load image - sphere in 64x64x64 array
sphere_array_size = 32
sphere_radius = 12
my_slice = [0,0.5]
# import pymrt as mrt
# import pymrt.geometry
# im = mrt.geometry.sphere(sphere_array_size,sphere_radius, 0.5)
# im = np.asarray(im)
# 3D imag
# noise_level = 0.5
# noise = noise_level*(np.max(im) - np.min(im))*np.random.rand(im.shape[0],im.shape[1],im.shape[2])

# Add noise:
# noisy_im = im + noise

noisy_im = imageio.imread('../tests/images/test3Dnoisy_sphere.tiff')

# Create mybifs BIFS object:
mybifs = bifs.BIFS()

# Can take a look at what functions and variables are available with, e.g.:
# dir(mybifs)

# Set a few things:
# Prior
mybifs.prior = "Gaussian" # Choices are currently: "Gaussian","Uniform"
# "Gaussian" is actually the default but for illustration...

# Lilelihood
mybifs.likelihood = "Gaussian" # Choices are currently: "Gaussian","Rician" 
# "Gaussian" is actually the default but again for illustration...

# Parameter Space Function
mybifs.param_func_type = "Inverse Power Decay"
# Current choices are: "Inverse Power Decay","Banded Inverse Power Decay",
# "Linear Decay" with default "Inverse Power Decay",
# but again for illustration...

# Can check comments in bifs.py for description of other parametere to set

# Load the image - note, typically just start here re. loading a noisy image
mybifs.load_image(noisy_im)

# Run BIFS
mybifs.BIFS_MAP()

# Take a look at the parameter function
# NOTE: For a reason I don't yet understand
# running this and/or the following utility
# functions before running mybifs.BIFS_MAP()
# causes mybifs.BIFS_MAP() to crash horribly
# with a string of errors like:
#
# Break on __THE_PROCESS_HAS_FORKED_AND_YOU_CANNOT_USE_THIS_
# COREFOUNDATION_FUNCTIONALITY___YOU_MUST_EXEC__() to debug.
# The process has forked and you cannot use this CoreFoundation
# functionality safely. You MUST exec().
#
bu.plot_param_func(mybifs)

# Look at the prior, liklelihood, and posterior at a voxel
bu.voxel_dist(mybifs,[mybifs.mod_image().shape[0]//2,mybifs.mod_image().shape[1]//2,mybifs.mod_image().shape[2]//2],do_plots=True)
  
# Plot the resulting image
# Get slices from 3D data
slice_index = np.int(np.round(my_slice[1]*mybifs.init_image().shape[my_slice[0]]))
if my_slice[0] == 0:
    init_im_slice = mybifs.init_image()[slice_index,:,:]
    init_mod_im_slice = mybifs.mod_image()[slice_index,:,:]
    fin_im_slice = mybifs.final_image()[slice_index,:,:]
    fin_mod_im_slice = mybifs.bifsk_image()[slice_index,:,:]
elif my_slice[0] == 1:
    init_im_slice = mybifs.init_image()[:,slice_index,:]
    init_mod_im_slice = mybifs.mod_image()[:,slice_index,:]
    fin_im_slice = mybifs.final_image()[:,slice_index,:]
    fin_mod_im_slice = mybifs.bifsk_image()[:,slice_index,:]
elif my_slice[0] == 2:
    init_im_slice = mybifs.init_image()[:,:,slice_index]
    init_mod_im_slice = mybifs.mod_image()[:,:,slice_index]
    fin_im_slice = mybifs.final_image()[:,:,slice_index]
    fin_mod_im_slice = mybifs.bifsk_image()[:,:,slice_index]
else:
    print("Sorry slice index needs to be one of 0,1,2") 

plt.subplot(221)
plt.axis('off')
plt.title("Initial Image")
plt.imshow(init_im_slice, cmap = cm.Greys_r)
  
# Initial K-Space Image
plt.subplot(223)
plt.axis('off')
plt.title("Initial K-Space Image")
showim1k = np.roll(np.roll(init_mod_im_slice,init_mod_im_slice.shape[0]//2,0),init_mod_im_slice.shape[1]//2,1)
plt.imshow(np.log(showim1k), cmap = cm.Greys_r)
  
# Final K-Space Image after running BIFS
plt.subplot(224)
plt.axis('off')
plt.title("Final K-Space Image")
showim2k = np.roll(np.roll(fin_mod_im_slice,fin_mod_im_slice.shape[0]//2,0),fin_mod_im_slice.shape[1]//2,1)
plt.imshow(np.log(showim2k), cmap = cm.Greys_r)

# Final Image after running BIFS
plt.subplot(222)
plt.axis('off')
plt.title("Reconstructed Image")
plt.imshow(fin_im_slice, cmap = cm.Greys_r)

plt.show()



###                  Utilities for BIFS                     ###
###                                                         ###
###  A set of utility functions for reading a jsonpickled   ###
###  BIFS object, examing the components of a BIFS object   ###
###  (e.g. image, transform representation, likelihood and  ###
###  prior at a voxel) during various stages                ###
###  of a BIFS analysis.                                    ###

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import scipy.stats as stats
from scipy.optimize import minimize_scalar as ms
import jsonpickle
import sys

def read_bifs(filename):
    bifs_handle = open(filename,"r")
    bifsj = bifs_handle.read()
    bifs_handle.close()
    return(jsonpickle.decode(bifsj))
 
def voxel_dist(bifs_obj,vox_coord, ostr=sys.stdout, do_plots=False):
    """
    voxel_dist() is a function for examining particular 
    choices of prior, likelihood, and posterior distributions at 
    a particular transform space point. It uses a Gaussian prior
    and focuses on illustrating the differences between the results
    of using the analytic form for a conjugate Gaussian likelihood and
    prior, or just directly calculating, via optimaization, use of a
    Gaussian or Rician likelihood. It outputs a number of parameters
    (e.g. distribution means and MAP estimates) and optionally plots
    the distributions, and assumes that an image has been loaded.
    
    Inputs:

    bifs_obj - a bifs object containing the Fourier representation
               of an image (i.e. an image has to have been loaded)

    vox_coord - an array containing the 1D, 2D, or 3D coordinates
                of a transform space voxel at which you want to
                examine the the prior, likelihood, and posterior
                distributions and some of their properties (e.g. mean)

    ostr      - optional output stream, stdout by default
                Textual output goes here

    do_plots  - optional boolean flag, False by default, determining whether
                or not to plot the distributions

    Output:

    Prints to ostr
               Prior mean
               Prior scale
               K-space data value
               Likelihood scale,
               Maximum of Rice Likelihood
               Maximum of Gaussian Likelihood, 
               Conjugate Gaussian value, 
               Gaussian Likelihood MAP estimate,
               Rician Likelihood MAP estimate
               k-space co-ordinate

    Optionally plots - Prior Distribution,
                       Gaussian Likelihood Distribution,
                       Rician Likelihood Distribution,
                       Posterior with Gaussian Likelihood,
                       Posterior with Rician Likelihood
        
    """
    # Normalization options:
    # 1. Match power in prior to power in image
    # 2. No normalization
    # norm = (np.sum(mybifs.mod_image()))/(np.sum(mybifs.prior_object(invalidate=False).mean()))
    norm = 1.0
        
    myPrior = bifs_obj.prior_object(invalidate=False)
    prior_mn = norm*myPrior.mean()[tuple(vox_coord)]
    prior_sc = norm*myPrior.sd()[tuple(vox_coord)]
    data_mn = bifs_obj.mod_image()[tuple(vox_coord)]
    data_sc = bifs_obj.likelihood_scale*data_mn
    
    # Shape parmeter for Rician
    b = 1.0/bifs_obj.likelihood_scale
    
    # Guassian conjugate prior case
    ksd2 = data_sc**2
    priorsd2 = prior_sc**2
    gauss_guass = ((prior_mn/priorsd2+data_mn/ksd2)/(1./priorsd2 + 1./ksd2))
        
    err_code = 0
    if np.size(vox_coord) != bifs_obj.imdim:
        raise BifsBadInputs("Co-ordinate dimension: {} is not the same as image dimension: {}".format(np.size(vox_coord), bifs_obj.imdim))
    else:
        print("Prior mean: ",prior_mn, file=ostr)
        print("Prior scale: ",prior_sc, file=ostr)
        print("K-space data value:",data_mn, file=ostr)
        print("Likelihood scale:",data_sc, file=ostr)

        # Distributions
        glike = lambda x: stats.norm.pdf(x,loc=data_mn,scale=data_sc)
        rlike = lambda x: stats.rice.pdf(x,b,scale=data_sc)
        prior = lambda x: stats.norm.pdf(x,loc=prior_mn,scale=prior_sc)
            
        findmin = lambda x: -rlike(x)
        res = ms(findmin, method='brent')
        print("Maximum of Rice Likelihood at: ",res.x, file=ostr)

        findmin = lambda x: -glike(x)
        res = ms(findmin, method='brent')
        print("Maximum of Gaussian Likelihood at: ",res.x, file=ostr)

        gposterior = lambda x: prior(x)*glike(x)
        rposterior = lambda x: prior(x)*rlike(x)

        print("Conjugate Gaussian value: ",gauss_guass, file=ostr)

        findming = lambda x: -gposterior(x)
        resg = ms(findming, method='brent')
        print("Gaussian Likelihood MAP estimate:",resg.x,"at k-space co-ordinates:",tuple(vox_coord), file=ostr)

        findminr = lambda x: -rposterior(x)
        resr = ms(findminr, method='brent')
        print("Rician Likelihood MAP estimate:",resr.x,"at k-space co-ordinates:",tuple(vox_coord), file=ostr)
        
        if do_plots:
            n_points = 100 # points to plot
            
            plot_start = prior_mn - 4.*prior_sc
            plot_end = prior_mn + 4.*prior_sc
            params_prior = np.linspace(plot_start,plot_end, n_points)
            plot_prior = np.array([prior(p) for p in params_prior])
            
            plot_start = data_mn - 4.*data_sc
            plot_end = data_mn + 4.*data_sc
            params_glike = np.linspace(plot_start,plot_end, n_points)
            plot_glike = np.array([glike(p) for p in params_glike])
            
            params_rlike = np.linspace(stats.rice.ppf(0.01,b,scale=data_sc),stats.rice.ppf(0.99,b,scale=data_sc), 100)
            plot_rlike = np.array([rlike(p) for p in params_rlike])
                
            plot_start = data_mn - 4.*data_sc
            plot_end = data_mn + 4.*data_sc
            params_glike_post = np.linspace(plot_start,plot_end, n_points)
            plot_glike_post = np.array([gposterior(p) for p in params_glike_post])
    
            plot_start = data_mn - 4.*data_sc
            plot_end = data_mn + 4.*data_sc
            params_rlike_post = np.linspace(plot_start,plot_end, n_points)
            plot_rlike_post = np.array([rposterior(p) for p in params_rlike_post])

            fig, axes = plt.subplots(5, 1, sharex=True, figsize=(8,8))
            axes[0].plot(params_prior, plot_prior)
            axes[0].set_title("Prior Distribution")
            axes[1].plot(params_glike, plot_glike)
            axes[1].set_title("Gaussian Likelihood Distribution")
            axes[2].plot(params_rlike, plot_rlike)
            axes[2].set_title("Rician Likelihood Distribution")
            axes[3].plot(params_glike_post, plot_glike_post)
            axes[3].set_title("Posterior with Gaussian Likelihood")
            axes[4].plot(params_rlike_post, plot_rlike_post)
            axes[4].set_title("Posterior with Rician Likelihood")
            plt.tight_layout()
            plt.show()
            plt.close("all")
    return err_code

def plot_param_func(bifs_obj):
    """
    plot_param_func() plots the current transform space parameter 
    function given a bifs object.

    Inputs:

    bifs_obj - a bifs object containing the Fourier representation
               of an image (i.e. an image has to have been loaded)

    Outputs:

    Plots current paramter function

    """
    prior = bifs_obj.prior_object(invalidate=False)
    im_size = bifs_obj.init_image().shape
    if bifs_obj.imdim == 1 or bifs_obj.imdim == 3:
        plt.title("K Space Parameter Function") 
        plt.xlabel("kx")
        KX = np.arange(im_size[0])
        # For now have to do the following by hand
        # Should figure some way to automate
        # when new parameter functions are added
        if bifs_obj.param_func_type == "Inverse Power Decay":
            Z = bifs_obj.bas.ixsc(prior.bvec(), KX, prior.decay())
        elif bifs_obj.param_func_type == "Banded Inverse Power Decay":
            Z = bifs_obj.bas.ixscbanded(prior.bvec(), KX, prior.decay(), prior.banded_cutoff())
        elif bifs_obj.param_func_type == "Linear Decay":
            Z = bifs_obj.bas.linsc(prior.bvec(), KX) 
        else:
            pass
        if bifs_obj.bumps:
            kdist = bifs_obj.kdist()
            Z += bifs_obj.bas.add_bumps(prior.bvec(), kdist, prior.bumps(), np.max(kdist))
        plt.plot(KX,Z)
    else: # Sort of silly since the parameter function is rotationally
          # symmetric (unless there's a trivial deviation due to kx != ky)
          # but it's pretty and might help check numerical glitches in filters
        fig = plt.figure()
        ax = plt.axes(projection='3d')
        # '3d' is not a documented projection type in https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.axes.html
        # execept that arbitrary strings are permitted.
        # previous code with
        #   ax = fig.gca(projection='3d')
        # Triggered this warning
        # j:\source\repos\bifs\bifs\bifs_util\util.py:210: MatplotlibDeprecationWarning: Calling gca() with keyword arguments was deprecated in Matplotlib 3.4.
        # Starting two minor releases later, gca() will take no keyword arguments.
        # The gca() function should only be used to get the current axes, or if no axes exist, create new axes with default keyword arguments.
        # To create a new axes with non-default arguments, use plt.axes() or plt.subplot().

        ax.set_title('K Space Parameter Function')
        ax.set_xlabel('kx')
        ax.set_ylabel('ky')
        KX = np.arange(im_size[0])
        KY = np.arange(im_size[1])
        KX, KY = np.meshgrid(KX, KY)
        Kdist = np.sqrt(KX**2 + KY**2)
        if bifs_obj.param_func_type == "Inverse Power Decay":
            Z = bifs_obj.bas.ixsc(prior.bvec(), Kdist, prior.decay())
        elif bifs_obj.param_func_type == "Banded Inverse Power Decay":
            Z = bifs_obj.bas.ixscbanded(prior.bvec(), Kdist, prior.decay(), prior.banded_cutoff())
        elif bifs_obj.param_func_type == "Linear Decay":
            Z = bifs_obj.bas.linsc(prior.bvec(), Kdist) 
        else:
            do_nothing = None
        if bifs_obj.bumps:
            kdist = bifs_obj.kdist()
            Z += bifs_obj.bas.add_bumps_to_pf(prior.bvec(), kdist, prior.bumps(), np.max(kdist))
        ZRoll = np.roll(np.roll(Z,(Z.shape[0]//2),0),(Z.shape[1]//2),1)
        surf = ax.plot_surface(KX, KY, ZRoll, cmap=cm.coolwarm,linewidth=0, antialiased=False)

    plt.show()
    plt.close('all')
    return



## Interface All Priors Must Implement

import numpy as np

class AbstractPrior:
    """A Prior holds the mean and variance of a prior to use in a particular analysis.
    At the moment, it is always Gaussian.
    Priors may apply various transformations to the original data, and will
    generally include parameters controlling those transformations.

    Returned mean, sd and variance should be numpy arrays with the same dimensions
    as the basis image being fit.

    Currently priors do NOT evaluate the likelihood of any data, though it would be a reasonable
    extension for them to do so.  This likelihood is just the likelihood of the data given the prior,
    not the overall Bayesion likelihood.

    Clients MUST only access these objects via their methods to assure proper caching
    and invalidation behavior when the parameters are changed.  An alternate strategy would be 
    to treat these objects as constant, with a new one created just before each calculation.  Given
    their potentially involved setup, that seems undesirable.
    """

    def distribution(self):
        "Type of distribution I represent"
        return "Gaussian"

    def mean(self):
        "return prior mean"

    def sd(self):
        "return prior standard deviation"

    def var(self):
        "return prior variance"
        s = self.sd()
        return s*s
    
    def name(self):
        "The human-friendly name of this prior"
        return str(type(self))

class AdjustmentPrior(AbstractPrior):
    """Performs some standard adjustments to the prior
        
    basis is, for now, a module, but the key thing is it must implement the 
        ixsc, ixscbanded, linsc, the corresponding bvec_* constants and add_bumps_to_pf.

    bvec - 2D float array specifying intercept and amplitude for parameter
           space functions

    scale - the overall scale of the prior variance
    scale_origin - prior scale at the origin - generally set huge
                           to allow data to determine overall scale
    bumps - dictionary containing set of "bump" filters to implement
    bump_types - set of choices for "bump" filter types to add to k-space
                    paramter function; uses scipy.signal window types 
                    so consult that documentation for available types - 
                    currently only types that only require window type name
                    and size are used - current choices are: 
                    "boxcar"
                    "blackman"
                    "hann"
                    "bartlett"
                    "flattop"
                    "parzen"
                    "bohman"
                    "blackmanharris"
                    "nuttall"
                    "barthann"
    bump_default_type - the default window type used (currently "blackman")

    Subclasses must provide methods
    _markDirty()
    mean() which should call _adjustMean()
    sd()

    And should set the instance variable
    _iorigin
    early on.  It is index of appropriate type and dimension to origin.

    """


    # These are class variables
    # Currently bumps are tied to the basis and only make sense for fourier
    bump_types = ["boxcar","blackman","hann","bartlett","flattop","parzen","bohman","blackmanharris","nuttall","barthann"]
    bump_default_type = "blackman"

    def __init__(self, basis, bvec, scale, scale_origin):
        self._bvec = bvec
        self._scale = scale
        self._scale_origin = scale_origin

        # Create an empty bump dictionary
        # keys are text strings for scipy.signal.window
        # available bump types from scipy.signal.window are 
        # (others like gaussian require more paramters):
        #
        # Due to our use in k-space picking the preferred versions
        # of these, which are the versions that end at 0
        # boxcar - kspace ring filter 
        # triang - pretty self explanatory - XXX - use bartlett instead 
        # blackman - nice decaying tails <- use as default
        # hamming - nice decaying tails - XXX
        # hann - nice decaying tails <- use this instead of hamming
        # bartlett - same as triang but ending at 0
        # flattop - fast decaying to below 0 and back <- should be 2nd def.
        # parzen - nice decaying tails
        # bohman - nice decaying tails
        # blackmanharris - good narrow, decaying tails 
        # nuttall - good narrow, decaying tail
        # barthann - kind of fat tails with pointy top
        #
        # Read about the properties of these at:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.get_window.html#scipy.signal.windows.get_window
        # functions and values are 3 element arrays containing:
        # 1) the position of the k-space filter in terms
        #    of the fraction of Kx max
        # 2) the amplitude of the k-space filter in terms
        #    of the fraction of parameter function max - bvec[1]
        # 3) the width of the filter to send to the window function
        #    specified in terms of a fraction of Kx max
        self._bumps = {}
        self._bump_type = type(self).bump_default_type
        self._basis = basis

    def _markDirty(self):
        "invalidate all cached values"


    def setbvec(self, bs):
        self._bvec = bs
        self._markDirty()

    def setScale(self, s):
        self._scale = s
        self._markDirty()

    def setScale_origin(self, s):
        self._scale_origin = s
        self._markDirty()

    def setbump_type(self, bt, reset=True):
        "Set a new bump type and empty existing bumps unless reset==False"
        # reset is ignored if bump type is unchanged
        assert bt in bump_types
        if bt != self._bump_type:
            self._bump_type = bt
            if reset:
                self._bumps = {}
            self._markDirty()

    def bvec(self):
        return self._bvec

    def scale(self):
        return self._scale

    def scale_origin(self):
        return self._scale_origin

    def bump_type(self):
        return self._bump_type

    def bumps(self):
        "consider this read only access"
        return self._bumps

    def add_bump(self,my_key,position,amplitude,width):
        """

        adds a bump filter to the self._bumps dictionary

        Inputs:
            my_key - the name of an appropriate scipy.signal shape
            position -  fraction of kmax for location of bump filter
            amplitude - fraction of maximum of parameter function 
                        for amplitudeof of bump filter 
            width -     fraction of kmax for width of bump filter        

        Outputs:
            adds bump filter to self.bumps dictionary

        """
        key_orig = my_key.split(".")[0]
        if my_key in self._bumps:
            isGoodKey = False
            for i in range(10): # No more than ten bump filters of any kind !!!!
                new_key = key_orig+"."+str(i)
                if new_key not in self._bumps:
                    isGoodKey = True
                    break
            if not isGoodKey:
                raise RuntimeError("No more than 10 bump filters of a type allowed.  Exceeded for type "+key_orig)
        else:
            new_key = my_key
        self._bumps[new_key] = np.array([float(position),float(amplitude),float(width)])
        self._markDirty()
        return

    def _adjustMean(self, mean):
        "return mean with all adjustments applied"
        if self._bumps:
            mean += self._basis.add_bumps_to_pf(self._bvec, self._kdist, self._bumps, np.int(np.max(self._kdist)))
        self._mean_init = mean[self._iorigin]
        mean[self._iorigin] = 0.
        return mean





###                        BIFS class                   ### 
###                                                     ###
###         Class for performing Bayesian Image         ###
###         Restoration in Fourier Space (BIFS)         ###

## The filename is bifscore, not bifs or BIFS to avoid trouble
## with the import machinery.  In particular, import bifs, at
## least when executed in this directory, could import bifs.py 
## (old name for this file) rather than the package.  And thus
## from bifs.priors would fail with an error that bifs was not 
## a package.

import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from pylab import *
import multiprocessing as mp
from scipy import stats
import imageio
from scipy.optimize import minimize_scalar as ms
from multiprocessing import Pool, TimeoutError
from datetime import datetime
import jsonpickle as jsp
from bifs.priors import FunctionalPrior as FP
from bifs.priors import EmpiricalPrior as EP
from bifs.bifsexception import *

from PyQt5.QtCore import QObject, Signal

import copy

class BIFS(QObject):
    """
    Class BIFS for performing Bayesian image analysis
    in fourier space

    Variables
    ---------

    In the next group, clients should always use the accessor functions listed:

    init_image() - initial loaded image
    k_image() - initial k-space image
    mod_image() - initial modulus image in k-space
    phase_image() - initial phase image in k-space
    bifsk_image() - final BIFS modulus image in k-space
    final_image() - final reconstructed image in image space

    NOTE: We're keeping all these images in the BIFS object for
    now re. testing and experimentation - might be useful to 
    eventually have production run options that are more
    parsimonius re. storage.

    image_file_loaded - whether an image is loaded (True,False)
    initial_image_file_name - file name of initial image
    imdim - int image dimension (1,2 or 3)

    kdist() = distance on the shifted k-space lattice

    view3Dslice - for 3D data this is a 2D array [a,b] where:
                  a = axis perpindicular to slice
                  b = fraction of maximum along that direction 
                      for slice location
    
    prior - string specifying prior distribution to use
            current choices:
            'Gaussian'

    prior_choices - list of current prior choices (see above)
    prior_scale - the overall scale of the prior variance
    prior_scale_orig - prior scale at the origin - generally set huge
                           to allow data to determine overall scale

    All of the above prior* variables are obsolescent.  Prefer the new
    _prior  - <AbstractPrior>

    The next 2 items might be obsolescent.  However, their presence permits us to
    set the parameter type before the image is loaded; _prior needs image info,
    specifically dimensions, for construction.
    param_func_type is replaced by _prior, specifically _prior.name(), 
    and param_func_choices should come from import priors
    param_func_type - string specifying the k-space BIFS paramter
                      function to use
                      current choices:
                      "Inverse Power Decay"
                      "Banded Inverse Power Decay"
                      "Linear Decay"
                      "Empirical"
    param_func_choices - list of current choices (see above)


    likelihood - string specifying likelihood distribution to use
                 current choices:
                 'Gaussian'
                 'Rician'

    likelihood_choices - list of current choices (see above)
    likelihood_scale - the assumed (const) noise level in k-space

    bessel_approx_lims - limits for bessel approximtion for rice
                         distribution - see paper referenced in code

    bessel_approx_array - array for bessel approximtion for rice
                         distribution - see paper referenced in code
    
    rice_denom_cutoff - cutoff for the demoninator of the closed form 
                        of the posterior with a Gaussian prior and
                        Rician likelihood derived from bessel approximation
                        see paper referenced in code

    basis - string specifying the basis to use - currently ony choice
            is "Fourier"
    basis_choices - list of current choices (see above)

    We continue to use bumps, but they mostly live in the prior.  We retain some info 
    here for compatibility with the GUI.

    bumps - dictionary containing set of "bump" filters to implement
    bump_types - set of choices for "bump" filter types to add to k-space
                 paramter function; uses scipy.signal window types 
                 so consult that documentation for available types - 
                 currently only types that only require window type name
                 and size are used - current choices are: 
                 "boxcar"
                 "blackman"
                 "hann"
                 "bartlett"
                 "flattop"
                 "parzen"
                 "bohman"
                 "blackmanharris"
                 "nuttall"
                 "barthann"
    bump_default_type - the default window type used (currently "blackman")


    Signals
    -------
    These are Qt signals, the reason the class must inherit from QObject.

    image_loaded - initial image has been loaded
    image_unloaded -  image has been unloaded

    """

    image_loaded = Signal()
    image_unloaded = Signal()
    
    def __init__(self,prior = "Gaussian",likelihood = "Gaussian",likelihood_scale = 0.05,prior_scale=0.0005,basis="Fourier"):
        """ 
       
        initializes class

        Inputs:
          prior - function type for prior
          likelihood - function type for likelihood
          likelihood_scale_orig - likelihood scale at k-space origin
          prior_scale - prior scale at k-space origin
          basis - basis type for transfrom space 

        Outputs:
          initializes class

        """

        super().__init__()

        self._invalidate_initial_image()
        self.view3Dslice = [0,0.5]

        self.prior = prior
        self.prior_choices = ["Gaussian"]
        self.prior_scale = prior_scale
        self.prior_scale_orig = 10.**7
        self._prior = None  # will hold <AbstractPrior> object
        self.param_func_choices = FP.param_func_choices()

        self.likelihood = likelihood
        self.likelihood_choices = ["Gaussian","Rician"]
        self.likelihood_scale = likelihood_scale
        self.rice_denom_cutoff = 0.0002
        
        # The following 3 parameters are used in conjuction
        # with a closed form for the posterior MAP estimate
        # with a Gaussian prior and Rician likelihood obtained
        # using sympy and based on an approximation to the first
        # order modified Bessel function discussed in:
        # A Simple and Efficient Approximation to the Modified Bessel
        # Functions and Its Applications to Rician Fading
        #     Ramy Salahat, Ehab Salahat, Ali Hakam and Tasneem Assaf
        # 2013 IEEE GCC Conference and exhibition, November 17-20, Doha, Qatar
        #
        self.bessel_approx_lims = np.array([0.0,11.5,20.0,37.5])
        self.bessel_approx_array  = np.array([[0.7536,0.4710,0.9807,1.144],
                                              [0.9739,-163.40,0.8672,0.995],
                                              [-0.715,0.9852,1.0795,0.5686],
                                              [0.2343,0.8554,1.0385,0.946]])
        self.basis_choices = ["Fourier"]
        self.basis = basis
        self.bas = None
        self.bumps = {}
        # Expand the following to if/else when more basis choices added
        # For now set default parameters here; make them editable via
        # pop widget
        if self.basis == "Fourier":
            from bifs.bases import fourier as fb
            self.bas = fb
        else:
            raise BifsBadInputs("Unknown basis: {}".format(self.basis))

        self.param_func_type = self.bas.param_func_type
        ##### Test 1 ##############
        ##### - These are the 'ideal' settings for a Rician
        ##### likelihood analysis
        # self.rice_denom_cutoff = 0.00025
        # self.likelihood = "Rician"
        # self.likelihood_scale = 0.0005
        # self.prior_scale = 0.00009
        ##### Test 1 ###############
        
    def _invalidate_initial_image(self):
        "Existing image, if any, and all that depends on it, is obsolete"
        self._init_image = None
        self.imdim = None
        self.image_file_loaded = False
        self.image_unloaded.emit()
        self.initial_image_file_name = ''
        self._prior = None  # depends on image dimensions
        self._invalidate_kspace()

    def _invalidate_kspace(self):
        """
        The transformation to k-space has or will change.
        Everything that depends on it is invalid
        """
        self._kdist = None
        self._k_image = None
        self._mod_image = None
        self._phase_image = None
        self._invalidate_final()

    def _invalidate_final(self):
        """
        The mapping from the k-space image to the final image has changed.
        Invalidate everything that depends on it.
        """
        self._final_image = None
        self._bifsk_image = None

    def save_results(self):
        """

        pickles current bifs object and saves to a time stamped file
        using jsonpickle; in addition saves processed final image and 
        final transform space image to files with the same time stamp 
        in case user just wants copies of final images without having 
        to reload bifs object.

        Inputs:

          current bifs object
         
        Outputs:
          saves time stamped files containing images and bifs parameters

        """
        # Date stamp for parameter output file
        date_stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        # Output file names
        if self.image_file_loaded:
            # Assume here that input file is in standard 2 section,
            # period seperated form with file suffix as image type
            # First strip off any possible preceeding directory names
            in_file_name = self.initial_image_file_name
            in_file = in_file_name.split('/')[-1]
            file_prefix = in_file.split('.')[0]
            image_type = in_file.split('.')[1]
        else:
            # file_prefix = 'bifs_np_array_input'
            in_file_name = "Numpy Array"
            image_type = 'bmp'
            
        # Parameter file:
        out_p = 'bifs_params_'+date_stamp+'.bifspout'
        out_im = 'bifs_output_image_'+date_stamp+'.'+image_type
        out_k = 'bifs_output_kspace_'+date_stamp+'.'+image_type
        # Center final K-space image
        out_k_im = np.roll(np.roll(self.bifsk_image(),self.bifsk_image().shape[0]//2+1,0),self.bifsk_image().shape[1]//2,1)
        # Output images
        plt.imsave(out_im,self.final_image(),cmap = cm.Greys_r)
        plt.imsave(out_k,np.log(out_k_im),cmap = cm.Greys_r)
        # Output paramter file
        param_file = open(out_p, "w")
        param_file.write(jsp.encode(self))
        param_file.close()
        return

    def copy_params(self):
        """

        Return a new bifs object that has all the basic parameter values found in self.

        Does not copy the image or filename
        """
        newbifs = BIFS()
        newbifs.param_func_type = self.param_func_type
        newbifs.prior = self.prior
        newbifs.prior_scale = self.prior_scale
        newbifs._prior = copy.deepcopy(self._prior)
        newbifs.likelihood = self.likelihood
        newbifs.likelihood_scale = self.likelihood_scale
        newbifs.bumps = self.bumps
        newbifs.view3Dslice = self.view3Dslice
        return newbifs


    def load_image_file(self,fileName):
        """
        
        load an image file using name fileName; can use to
        examine image without invoking full bifs initialization.
        
        Inputs:
          fileName - name of file to load

        Outputs:
          loads file into bifs object

        RB makes read_imfile persistent in some cases so bulk scans can get the headers.

        """
        self._invalidate_initial_image()
        self.initial_image_file_name = fileName
        try:
            # For now the convention is that 1D data sets
            # will have filennames ending in '.txt' and
            # will be in numpy text file form, e.g. as created
            # with numpy.savetxt with the conventions that
            # header comments start with # and there is no
            # sperator between lines
            if self.initial_image_file_name.lower().endswith('.txt'):
                # numpy.loadtext automatacilly reads to numpy array
                read_image = np.loadtxt(self.initial_image_file_name)
            else:
                try:
                    import nibabel
                    # if nibabel is not installed, move on
                    # if nibabel can't read file, move on
                    # nibabel.streamlines.is_supported(fname) would seem to be a good test, but files
                    # that fail it can still be read, e.g., nii.gz files.
                    # So we use the Python idiom "Better to ask forgiveness than permission"
                    self.read_imfile = nibabel.load(self.initial_image_file_name)
                    read_image = self.read_imfile.get_fdata()
                except:
                    try:
                        self.read_imfile = imageio.volread(self.initial_image_file_name)
                        assert len(self.read_imfile) > 2
                        read_image = np.asarray(self.read_imfile)
                    except:
                        # we have a 2D, or maybe 1D, image
                        self.read_imfile = imageio.imread(self.initial_image_file_name)
                        read_image = np.asarray(self.read_imfile)
            self.load_image(read_image, invalidate=False)
        except Exception as exc:
            raise BifsBadInputs(f"BIFS Couldn't read image file: {fileName}") from exc
        return

    def load_image(self, init_image, invalidate=True):
        """

        intializes the bifs object so as to be ready for
        analysis steps.

        load_image_file calls this

        Inputs:

           init_image - array generated by loading image file using
                       load_image_file()
           invalidate - if True, clear caches.  Only very knowledgeable clients should use
                        this set to False

        """
        if invalidate:
            self._invalidate_initial_image()
        self._init_image = init_image
        self._init_image[np.isnan(init_image)] = 0.0
        self.imdim = len(init_image.shape)
        if not self.imdim in [1, 2, 3]:
            raise BifsBadInputs(f"Images with {self.imdim} dimensions are not supported.  Must be 1, 2, or 3 dimensions.")
        self.image_file_loaded = True
        self.image_loaded.emit()
        return

    def init_image(self):
        "Return initial image, if any"
        # For consistency with other accessors
        # if it's not there, we can't do anything
        return self._init_image

    def kdist(self):
        if self._kdist is None:
            myShape = self._init_image.shape
            if self.imdim == 1:
                self._kdist = self.bas.kdist1D(*myShape)
            elif self.imdim == 2:
                self._kdist = self.bas.kdist2D(*myShape)
            elif self.imdim == 3:
                self._kdist = self.bas.kdist3D(*myShape)
        return self._kdist

    def k_image(self):
        if self._k_image is None:
            if self.imdim == 1:
                self._k_image = self.bas.tx1(self._init_image) # Get k-space image
            elif self.imdim == 2:
                self._k_image = self.bas.tx2(self._init_image) # Get k-space image
            elif self.imdim == 3:
                self._k_image = self.bas.txn(self._init_image) # Get k-space image
            else:
                raise BifsBadInputs(f"1, 2, or 3 dimensional images supported. {self.imdim} dimensions are not supported." )
        return self._k_image

    def _final_setup(self):
        """Setup after image loaded and all other parameters are set.
        This used to be part of load_image, but was vulnerable to
        making calculations based on parameters that would later change.

        This should mostly be delegated to a suitable basis object.
        """
        
        if self.basis == "Fourier": # Add other basis functions as else...
            self._mod_image = abs(self.k_image()) # Get modulus image in k-space
            # sp.angle is deprecated
            self._phase_image = np.angle(self.k_image()) # Get phase image in k-space
            # self.data_std =  self.likelihood_scale*self.mod_image
            self.image_exists = True
        else:
            raise BifsBadInputs(f"{self.basis} basis unsupported.  Use Fourier.")

        # Set prior via orthogonal-space parameter function
        self.set_prior_func_type(self.param_func_type)
        # Set Rice parameter in case you use it
        # self.rice_arg = self.mod_image/self.likelihood_scale


        #### - The following doesn't seem to work, i.e. de-emphasizes -  ####
        #### - prior to too large an extent - need to investigate this - ####
        # Do normalization
        # # Since modulus and parameter functions are positive definite
        # don't need to square them (i.e. k-space integral) to get power
        # self.norm = (np.sum(self.mod_image))/(np.sum(self.prior_mean))
        # self.prior_mean = self.norm*self.prior_mean
        # self.prior_std = self.norm*self.prior_std
        # data std is regarded as constant, related to SNR
        # need to figure out best way to estimate and normalize
        ####                                                             ####
        ####                                                             ####
        
        return

    def mod_image(self):
        "Return modulus of k-space image"
        if self._mod_image is None:
            self._final_setup()
        return self._mod_image

    def phase_image(self):
        "return phase of k-space image"
        if self._phase_image is None:
            self._final_setup()
        return self._phase_image

    def final_image(self):
        "return final image in conventional space"
        if self._final_image is None:
            self.BIFS_MAP()
        return self._final_image

    def bifsk_image(self):
        "return final image in k-space"
        if self._bifsk_image is None:
            self.BIFS_MAP()
        return self._bifsk_image

    def add_bump(self,my_key,position,amplitude,width):
        """

        adds a bump filter to the self.bumps dictionalry

        Inputs:
           my_key - the name of an appropriate scipy.signal shape
           position -  fraction of kmax for location of bump filter
           amplitude - fraction of maximum of parameter function 
                       for amplitudeof of bump filter 
           width -     fraction of kmax for width of bump filter        

        Outputs:
          adds bump filter to prior

        """
        self._invalidate_final()
        return self.prior_object().add_bump(my_key, position, amplitude, width)


    def set_prior_func_type(self, pft : str):
        """ Set up prior object
        pft is the name of a type of prior function
        """
        if not self.image_file_loaded:
            self.param_func_type = pft
            self._prior = None
            # Nothing more can be done until image is loaded
            return
        if self._prior is not None and pft == self._prior.name():
            return
        self._invalidate_final()
        # test ##########
        # The following doesn't work and I don't know why
        # It seems that the prior function (in particulr the decay
        # function) should be normalized by the zero k-space value
        # of the image (i.e. the total "power") rather than an arbitrary
        # value like 500
        # self.bvec[1] = self.mod_image[0,0]
        # print("Zero k-space value:",self.mod_image[0,0])
        # self.bvec[1] = 50.0
        # test ##########
        if self.basis == "Fourier": # the usual re. elses for other tx spaces
            # Try setting bvec[1] = self.mod_image[0]
            # if self.imdim == 1:
            #    self.bvec[1] = self.mod_image[0]
            # elif self.imdim == 2:
            #    self.bvec[1] = self.mod_image[0,0]
            # elif self.imdim == 3:
            #    self.bvec[1] = self.mod_image[0,0,0]
            # else:
            #    pass
            if pft == "Inverse Power Decay":
                self._prior = FP.InversePowerDecayPrior(self.bas, self.kdist(), scale = self.prior_scale, scale_origin = self.prior_scale_orig)
            elif pft == "Banded Inverse Power Decay":
                self._prior = FP.BandedInversePowerDecayPrior(self.bas, self.kdist(), scale = self.prior_scale, scale_origin = self.prior_scale_orig)
            elif pft == "Linear Decay":
                self._prior = FP.LinearDecayPrior(self.bas, self.kdist(), scale = self.prior_scale, scale_origin = self.prior_scale_orig)
            elif pft == "Empirical":
                # This should already have been handled via set_empirical()
                assert self._prior.name() == "Empirical"
            else:
                raise BifsBadInputs("Please specify recognized transform space parameter function, one of:"+
                                   ", ".join(self.param_func_choices))
            self.param_func_type = pft
        else:
            raise BifBadInputs(f"{self.basis} basis unsupported.  Use Fourier.")

    def prior_object(self, invalidate=True):
        """ Return object representing the prior for possible editing.
        Edits may change the type-specific parameters of the object,
        but not the type (functional form) of the object.

        The ugly name is a result of prior already being used as a string.
        It should be the case that self.prior == self.prior_object().distribution().

        Also, the whole interface is ugly.

        Set invalidate=False only if you are not modifying the prior.
        """
        if invalidate:
            self._invalidate_final()
        if self._prior is None:
            self.set_prior_func_type(self.param_func_type)
        return self._prior

    def load_empirical(self, fname):
        """Load empirical prior from named file and set mode to Empirical
        """
        x = np.load(fname)
        return self.set_empirical(x)


    def set_empirical(self, x):
        """x is an object with the mean and sd of distn at each voxel
        It is presumed to be empirically derived, though you could make one up
        if you wanted to.

        Sets prior scale to 1, since the default value is very small.
        You can and probably should make it larger via the Gaussian gui specification.

        Note that because this requires self.bas and self.kdist the *image must be loaded first*.
        """
        self._invalidate_final()
        self._prior = EP.EmpiricalPrior(x, self.bas, self.kdist(), scale = 1.0, scale_origin = self.prior_scale_orig)
        self.param_func_type = "Empirical"

    def bifs_map_gauss_gauss(self, prior, mf, ds2):
        """
        returns MAP estimate for posterior function using 
        gaussian prior and likelihood and uses analytic 
        conjugate function

        Inputs:
           prior - <AbstractPrior>
           mf - k-space modulus (from data)
           ds2 - square of likelihood std

        Outputs:
           MAP estimate

        """
        # return((pm/ps2 + mf/ms2) / (1./ps2 + 1./ms2))
        # more efficient:
        return((prior.mean()*ds2 + mf*prior.var()) / (prior.var() + ds2))

    # Don't think the arguments to the prior and likelihood are
    # right yet, e.g. maybe need self.prior_mean_orig as argument
    # for scale factor of Gaussian prior - need to estimate the
    # the scale factor for the Rician from the signal to noise
    # in the image... KY
    
    def bifs_map_gauss_rice(self, prior, mf, ds):
        """
        returns MAP estimate for posterior function using 
        gaussian prior and rician likelihood and uses uses analytic
        function derived using sympy and the modified Bessel
        function approximation from the paper cited above
        
        Inputs:
           prior - <AbstractPrior>
           mf - k-space modulus (from data)
           ds - data std
        Outputs:
           MAP estimate
        """
        # Kludge - slows things down but we use the Gaussian conjugate
        # prior calculation to estimate where the MAP value will be and
        # ergo estimate which of the exponential coefficient values to use
        # for the modified Bessel function approximation
        #
        conj = self.bifs_map_gauss_gauss(prior, self.mod_image, self.ksd2)
        d = (2*ds**2 - 2*prior.var())
        dabs = np.abs(d)
        # Estimate which Bessel approximation coefficients
        # to use based on the Gaussian conjugate MAP value

        sn = np.zeros(d.shape)
        sn = where(np.abs(d) > 0,np.sign(d),-1.0)
        denom = sn*np.maximum(np.abs(d),self.rice_denom_cutoff)
        b = mf/ds
        result = 0.0
        for i in range(self.bsa.shape[-1]):
            if self.imdim == 1:
                ba = self.bsa[:,i]
            elif self.imdim == 2:
                ba = self.bsa[:,:,i]
            elif self.imdim == 3:
                ba = self.bsa[:,:,:,i]
            else:
                pass
            num = (-(b*ba*ds*prior.var() - mf*ds**2 + 2*mf*prior.var() - ds**2*prior.mean() + 
                     ds*np.sqrt(b**2*ba**2*prior.var()**2 + 2*b*ba*mf*ds*prior.var() - 2*b*ba*ds*prior.mean()*prior.var() +
                               mf**2*ds**2 - 2*mf*ds**2*prior.mean() + ds**2*prior.mean()**2 - 4*ds**2*prior.var() + 4*prior.var()**2)))
            ### Note - the closer the scales get to each other
            # (ergo the pathalogical blow up), the MAP estimate
            # should actually get closer to the average of the
            # the prior mean and likelihood mean - that's the
            # logic for:
            # denom = np.where(dabs < self.rice_denom_cutoff,(2*num)/(mf+pm),dabs)
            ### Doesn't seem to work so back to above denom setting with
            ### cutoff
                
            result += (num/denom)/len(ba.shape)
        # Need to include the sign part here in case the setting
        # for the denominator screwed up the sign
        return np.sign(result)*result
        
    def BIFS_MAP(self):
        """

        performs MAP optimization individually at each k-space 
        point using the MAP estimation functions above,
        recombines the results, performs inverse transform and returns final image.

        This seemed like a good place to use the multiprocessing package, but initial
        testing found that to be slower, and so we removed the code.

        Inputs:

        Outputs:
          Sets final_image based on performing k-space MAP estimation 

        """
        # Break up image(s) into linear array(s) for sending off
        # to multiprocessing - this is stupid and unecessarily
        # time consuming but I still need to figure out how to send
        # chunks of multivariate arrays to a multiprocessing pool
        if not self.image_file_loaded:
            raise BifsBadInputs("Error: Need to load an image before running MAP")
        self._final_setup()

        # Square of likelihood
        self.ksd2 = self.likelihood_scale**2
        # self.data_std2 = self.data_std**2
        # Rician parameter
        # self.rice_arg = self.mod_image/self.likelihood_scale
        if self.prior == "Gaussian" and self.likelihood == "Gaussian":
            self._bifsk_image = self.bifs_map_gauss_gauss(self._prior, self.mod_image(), self.ksd2)
        elif self.prior == "Gaussian" and self.likelihood == "Rician":
            conj = self.bifs_map_gauss_gauss(self._prior, self.mod_image(), self.ksd2)
            besind = np.zeros(self._init_image.shape, dtype=int)
            besind[np.where(conj > self.bessel_approx_lims[1])] = 1
            besind[np.where(conj > self.bessel_approx_lims[2])] = 2
            besind[np.where(conj > self.bessel_approx_lims[3])] = 3
            self.bsa = self.bessel_approx_array[besind,:]
            self._bifsk_image = self.bifs_map_gauss_rice(self._prior, self.mod_image(), self.likelihood_scale)
        else:
            raise BifsBadInputs(f"Unable to calculate with {self.prior} prior and {self.likelihood} likelihood.")
        # Send back to image space
        if self.basis == "Fourier": # usual add else for other tx 
            if self.imdim == 1:
                self._final_image = np.real(self.bas.itx1(self._bifsk_image*np.exp(1j*self.phase_image())))
            elif self.imdim == 2:
                self._final_image = np.real(self.bas.itx2(self._bifsk_image*np.exp(1j*self.phase_image())))
            elif self.imdim == 3:
                self._final_image = np.real(self.bas.itxn(self._bifsk_image*np.exp(1j*self.phase_image())))
        else:
            raise BifsBadInputs(f"{self.basis} basis unsupported.  Use Fourier.")
        return



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='bifs',
    description='Bayesian Imaging in Fourier Space (BIFS)',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='John Kornak, Karl Young, Ross Boylan',
    author_email='john.kornak@ucsf.edu,kyoung21b2000@gmail.com',
    maintainer_email='ross.boylan@ucsf.edu,bifs@ucsf.edu',
    url='https://github.com/ucsf-deb/bifs',
    license=license,
    packages=['bifs'],
    include_package_data=True,
    # current documentation only gives the setup.cfg syntax for entry points
    # and the translation is not obvious.  Used https://stackoverflow.com/questions/774824/explain-python-entry-points
    entry_points = {
        "gui_scripts" : ["bifs_gui = bifs.bifs_gui:launch"]
        },

    # remember to keep in sync with requirements.txt
    # nibabel is only necessary if your image format requires it
    install_requires=['imageio',
                      'jsonpickle',
                      'matplotlib',
                      'nibabel',
                      'numpy >= 1.17',
                      'PyQt5',
                      'scipy'],
    # not extras_requireS
    extras_require={"NI" : "nibabel",
                    "tests" : "pytest"},

    # license type TBD
    classifiers = ['Programming Language :: Python :: 3',
                   'Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'Natural Language :: English',
                   'Operating System :: OS Independent',
                   'Topic :: Multimedia :: Graphics',
                   'Topic :: Scientific/Engineering :: Image Processing',
                   'Topic :: Scientific/Engineering :: Visualization',
                   ]
)





###                 BIFS Fourier basis functions                ###
###                                                             ###
### Global variables:                                           ###
###                                                             ###
### param_func_choice - Current set of k-space                  ###
###                     function choices                        ###
###                                                             ###
### param_func_type - Current parameter funct type              ###
###                                                             ###
### bvec_ixsc - default offset and amplitude for                ###
###             power law decay function                        ###
###                                                             ###
### bvec_ixscbanded - default offset and amplitude for          ###
###                   banded power law decay function           ###
###                                                             ###
### bvec_linsc - default offset and amplitude for               ###
###              "linear" decay function                        ### 
###                                                             ###
### tx1 - bifs transform agnostic name for scipy.fftpack fft    ###
###                                                             ###
### itx1 - bifs transform agnostic name for scipy.fftpack ifft  ###
###                                                             ###
### tx2 - bifs transform agnostic name for scipy.fftpack fft2   ###
###                                                             ###
### itx2 - bifs transform agnostic name for scipy.fftpack ifft2 ###
###                                                             ###
### txn  - bifs transform agnostic name for scipy.fftpack fftn  ###
###                                                             ###
### itxn - bifs transform agnostic name for scipy.fftpack ifftn ###


import numpy as np
from scipy.fftpack import fft,ifft,fft2,ifft2,fftn,ifftn
from scipy import signal

# Choices for K-space parameter function
param_func_choices = ["Inverse Power Decay","Banded Inverse Power Decay","Linear Decay", "Empirical"]

# Default K-space parameter function choice
param_func_type = "Inverse Power Decay"

# Some defaults for parameter function parameters
bvec_ixsc = np.array([0.,500.])
bvec_ixscbanded = np.array([0.,500.])
bvec_linsc = np.array([0.1,500.])

# BIFS conformant transform names

tx1 = fft
itx1 = ifft
tx2 = fft2
itx2 = ifft2
txn = fftn
itxn = ifftn

# K-space distance calculations

def kdist_calc(N):
    """
    kdist(N)

    kdist outputs centered k-space indicex array

    Inpute:
    
    N - Length of desired k-space index array

    Outputs:

    centered, length N k-space index array

    """
    kval = np.zeros((1,N))
    kval[0,0:N//2] = range(1,N//2+1)
    kval[0,0:N//2] = kval[0,0:N//2] - 1 # try this so you get an origin 
    kval[0,N//2:N] = range(-N//2,0)
        
    return kval
        

def kdist1D(N1):
    """
    kdist1D() generates array with distance from
    center of Fourier space but shifted so origin is at index 0 

    Inpute:
    
    N1 - Size of desired array

    Outputs:

    shifted, length N1 k-space distance array

    """
    xvec = kdist_calc(N1)
    kdmat = np.zeros((N1))
    for i in range(N1):
        kdmat[i] = np.sqrt(xvec[0,i]**2)
    return kdmat
    
def kdist2D(N1,N2):
    """
    kdist2D() generates array with distance from
    center of Fourier space but shifted so origin is at index (0,0) 

    Inpute:
    
    N1 - Size of desired array in kx
    N2 - Size of desired array in ky

    Outputs:

    shifted, size (N1,N2) k-space distance array

    """
    xvec = kdist_calc(N1)
    yvec = kdist_calc(N2)
    kdmat = np.zeros((N1,N2))
    for i in range(N1):
        for j in range(N2):
            kdmat[i,j] = np.sqrt(xvec[0,i]**2 + yvec[0,j]**2)
    return kdmat
    
def kdist3D(*ns):
    """
    kdist3D() generates array with distance from
    center of Fourier space but shifted so origin is at index (0,0,0) 

    Input:

    ns should be 3 arguments:
    N1 - Size of desired array in kx
    N2 - Size of desired array in ky
    N3 - Size of desired array in kz 

    Outputs:

    shifted, size (N1,N2,N3) k-space distance array

    Comment: This code could almost work for the 1 and 2D cases,
    but doing so would require a way to generalize the
    a+b+c expression. It is rougly 400x faster than a naive loop-based implmentation.
    """
    dist = [kdist_calc(n)[0, ]**2 for n in ns]
    a, b, c = np.ix_(*dist)
    kdmat = np.sqrt(a+b+c)
    return kdmat


# K-space functional forms for parameter function 

def ixsc(bvec,x,y):
    """
    ixsc() generates a k-space parameter function for the modulus of 
    the prior mean of the form: 
                 amplitude/distance^y + offset
    where the offset is from center of k-space. 

    Inputs:

    bvec - 2 element array where bvec[0] is offset, bvec[1] is amplitude

    x - function argument

    y - power law decay index

    Outputs:

    function value

    """
    zero_offset = 0.01
    try:
        if np.abs(x) > 0.0:
            return bvec[0] + bvec[1]*x**(-y)
    except:
        return bvec[0] + bvec[1]*(x+zero_offset)**(-y)

      
def ixscbanded(bvec,x,y,z):
    """
    ixscbanded() generates a k-space parameter function for the modulus of 
    the prior mean of the form: 
                 amplitude/distance^y + offset
    where the offset is from center of k-space and in addition has a 
    cutoff at a distance z from the center of k-space.

    Inputs:

    bvec - 2 element array where bvec[0] is offset, bvec[1] is amplitude

    x - function argument

    y - power law decay index

    z - cutoff value, distance from center of k-space

    Outputs:

    function value

    """
    idx1 = x < z
    zero_offset = 0.01
    zero_floor = 0.0001
    try:
        if np.abs(x) > 0.0:
            idx2 = bvec[0] + bvec[1]*x**(-y)
    except:
        idx2 = bvec[0] + bvec[1]*(x+zero_offset)**(-y)
    return idx1*idx2 + zero_floor
    # return np.where(x<z,bvec[0] + bvec[1]*x**(-y),0.)

def linsc(bvec,x):
    """
    linsc() generates a k-space parameter function for the modulus of 
    the prior mean of the form: 
                 offset - amplitude*x^2
    where the offset is from center of k-space. 

    Inputs:

    bvec - 2 element array where bvec[0] is offset, bvec[1] is amplitude

    x - function argument

    Outputs:

    function value
    
    """
    zero_floor = 0.0001
    if not (bvec[0] > 0.0): # Assume there was a mistake like
                            # resetting param_func_type and
                            # forgetting to reset bvec - should
                            # probably fix this by using getter's
        bvec[0] = 0.1  # default bvec[0]
        bvec[1] = 500. # might as well set bvec[1] to default in this case
    y = bvec[1] - bvec[0]*x**2
    return np.where(y>0,y,zero_floor)

def add_bumps_to_pf(bvec,x,bumps,kmax):
    """ 
    add_bumps_to_pf() adds filter "bumps" to the k-space paramter
    function to enhance or supress certain frequency ranges in 
    generating the prior mean. Since the k-space paramter functions 
    currently used are rotationally symmetric this adds a ring or 
    spherical shell in 2D or 3D. This function uses the window shapes 
    from scipy.singal to generate the "bumps" so all filter shapes that 
    only require the filter type and size are available. The documentation
    for scipy.signal can be be consulted for a current list of available
    filter shapes.

    Inputs:
    
    bvec - 2 element array where bvec[0] is offset, bvec[1] is amplitude
           for current k-space paramter function; bvec[1] is used to
           generate the amplitude of the bump re. relative fraction of 
           bvec[1].

    x - argument for the bump function(s)

    bumps - a python dictionary with elements of the form:

            bump_name:[bump_location,bump_amplitude,bump_width]

            where:
            bump_name(key) - a text string containing one of the 
                             available scipy.signal filter forms
            bump_location -  location of the bump as fraction of the
                             maximum k-space value
            bump_amplitude - amplitude of the bump as fraction of the
                             maximum of the k-space paramter function,
                             i.e. bvec[1]
            bump_width -     width of the bump as fraction of the
                             maximum k-space value

    kmax - maximum k-space value

    Outputs:

    Function with all bumps in the bumps dictionary added (to be
    added to the paramter function)

    """

    # Make cumlative bump function to add
    
    bump_func = np.zeros(np.int(kmax))
    if bumps: # just to be sure
        bump_count = 0
        #
        # print(bumps)
        #
        for k,v in bumps.items():
            bump_center = np.int(v[0]*kmax)
            bump_amp = float(v[1]*bvec[1])
            bump_width = np.int(v[2]*kmax)
            start = bump_center - np.int(np.floor(bump_width/2))
            stop = bump_center + np.int(np.ceil(bump_width/2))
            #
            # print(x,kmax,k,v,bump_center,bump_amp,bump_width,start,stop)
            #
            # Split k in case multiple cases of same filter type
            # in which case the convention is to append ".something"
            bump_func[start:stop] = bump_amp*signal.get_window(k.split('.')[0],bump_width)
            if bump_count == 0:
                y = bump_func[np.asarray(x, dtype=int)-1]
            else:
                y += bump_func[np.asarray(x, dtype=int)-1]
            #
            # print("nonzero y in add_bumps",np.sum(np.where(y !=0,1,0)))
            #
        return np.where(y > 0.0,y,0.0)
#    except:
#        print("Couldn't implement bump function additions")
#        return
    



import bifs
import os
import numpy as np

def test_simple():
	b = bifs.BIFS()
	b.load_image_file("tests/images/lena512.bmp")
	b.BIFS_MAP()
	assert (b.prior_object().var() == b.prior_object().sd()**2).all()

def test_reset():
	b = bifs.BIFS()
	b.load_image_file("tests/images/lena512.bmp")
	b.BIFS_MAP()

	b2 = bifs.BIFS()
	b2.load_image_file("tests/images/lena512.bmp")
	b2.BIFS_MAP()
	b2.set_prior_func_type("Linear Decay")
	b2.BIFS_MAP()

	assert b.bas == b2.bas
	assert (b.prior_object().mean() != b2.prior_object().mean()).any()
	assert b.prior_object().name() != b2.prior_object().name()

def test_reset2():
	b = bifs.BIFS()
	b.load_image_file("tests/images/lena512.bmp")
	b.BIFS_MAP()

	b2 = bifs.BIFS()
	b2.set_prior_func_type("Linear Decay")
	b2.load_image_file("tests/images/lena512.bmp")
	b2.BIFS_MAP()


	assert b.bas == b2.bas
	assert (b.prior_object().mean() != b2.prior_object().mean()).any()
	assert b.prior_object().name() != b2.prior_object().name()

def test_inverse():
	b = bifs.BIFS()
	b.set_prior_func_type("Banded Inverse Power Decay")
	b.load_image_file("tests/images/lena512.bmp")
	b.BIFS_MAP()

